"""Provides HeAT's core data structure, the DNDarray, a distributed n-dimensional array"""
from __future__ import annotations

import math
import numpy as np
import torch
import warnings

from inspect import stack
from mpi4py import MPI
from pathlib import Path
from typing import List, Union, Tuple, TypeVar, Optional

warnings.simplefilter("always", ResourceWarning)

# NOTE: heat module imports need to be placed at the very end of the file to avoid cyclic dependencies
__all__ = ["DNDarray"]

Communication = TypeVar("Communication")


class LocalIndex:
    """
    Indexing class for local operations (primarily for :func:`lloc` function)
    For docs on ``__getitem__`` and ``__setitem__`` see :func:`lloc`
    """

    def __init__(self, obj):
        self.obj = obj

    def __getitem__(self, key):
        return self.obj[key]

    def __setitem__(self, key, value):
        self.obj[key] = value


class DNDarray:
    """
    Distributed N-Dimensional array. The core element of HeAT. It is composed of
    PyTorch tensors local to each process.

    Parameters
    ----------
    array : torch.Tensor
        Local array elements
    gshape : Tuple[int,...]
        The global shape of the array
    dtype : datatype
        The datatype of the array
    split : int or None
        The axis on which the array is divided between processes
    device : Device
        The device on which the local arrays are using (cpu or gpu)
    comm : Communication
        The communications object for sending and receiving data
    balanced: bool or None
        Describes whether the data are evenly distributed across processes.
        If this information is not available (``self.balanced is None``), it
        can be gathered via the :func:`is_balanced()` method (requires communication).
    """

    def __init__(
        self,
        array: torch.Tensor,
        gshape: Tuple[int, ...],
        dtype: datatype,
        split: Union[int, None],
        device: Device,
        comm: Communication,
        balanced: bool,
    ):
        self.__array = array
        self.__gshape = gshape
        self.__dtype = dtype
        self.__split = split
        self.__device = device
        self.__comm = comm
        self.__balanced = balanced
        self.__ishalo = False
        self.__halo_next = None
        self.__halo_prev = None
        self.__partitions_dict__ = None
        self.__lshape_map = None

        # check for inconsistencies between torch and heat devices
        assert str(array.device) == device.torch_device

    @property
    def balanced(self) -> bool:
        """
        Boolean value indicating if the DNDarray is balanced between the MPI processes
        """
        return self.__balanced

    @property
    def comm(self) -> Communication:
        """
        The :class:`~heat.core.communication.Communication` of the ``DNDarray``
        """
        return self.__comm

    @property
    def device(self) -> Device:
        """
        The :class:`~heat.core.devices.Device` of the ``DNDarray``
        """
        return self.__device

    @property
    def dtype(self) -> datatype:
        """
        The :class:`~heat.core.types.datatype` of the ``DNDarray``
        """
        return self.__dtype

    @property
    def gshape(self) -> Tuple:
        """
        Returns the global shape of the ``DNDarray`` across all processes
        """
        return self.__gshape

    @property
    def halo_next(self) -> torch.Tensor:
        """
        Returns the halo of the next process
        """
        return self.__halo_next

    @property
    def halo_prev(self) -> torch.Tensor:
        """
        Returns the halo of the previous process
        """
        return self.__halo_prev

    @property
    def larray(self) -> torch.Tensor:
        """
        Returns the underlying process-local ``torch.Tensor`` of the ``DNDarray``
        """
        return self.__array

    @larray.setter
    def larray(self, array: torch.Tensor):
        """
        Setter for ``self.larray``, the underlying local ``torch.Tensor`` of the ``DNDarray``.

        Parameters
        ----------
        array : torch.Tensor
            The new underlying local ``torch.tensor`` of the ``DNDarray``

        Warning
        -----------
        Please use this function with care, as it might corrupt/invalidate the metadata in the ``DNDarray`` instance.
        """
        # sanitize tensor input
        sanitation.sanitize_in_tensor(array)
        # verify consistency of tensor shape with global DNDarray
        sanitation.sanitize_lshape(self, array)
        # set balanced status
        split = self.split
        if split is not None and array.shape[split] != self.lshape[split]:
            self.__balanced = None
        self.__array = array

    @property
    def nbytes(self) -> int:
        """
        Returns the number of bytes consumed by the global tensor. Equivalent to property gnbytes.

        Note
        ------------
            Does not include memory consumed by non-element attributes of the ``DNDarray`` object.
        """
        return self.__array.element_size() * self.size

    @property
    def ndim(self) -> int:
        """
        Number of dimensions of the ``DNDarray``
        """
        return len(self.__gshape)

    @property
    def __partitioned__(self) -> dict:
        """
        This will return a dictionary containing information useful for working with the partitioned
        data. These items include the shape of the data on each process, the starting index of the data
        that a process has, the datatype of the data, the local devices, as well as the global
        partitioning scheme.

        An example of the output and shape is shown in :func:`ht.core.DNDarray.create_partition_interface <ht.core.DNDarray.create_partition_interface>`.

        Returns
        -------
        dictionary with the partition interface
        """
        if self.__partitions_dict__ is None:
            self.__partitions_dict__ = self.create_partition_interface()
        return self.__partitions_dict__

    @property
    def size(self) -> int:
        """
        Number of total elements of the ``DNDarray``
        """
        return (
            torch.prod(
                torch.tensor(self.gshape, dtype=torch.float64, device=self.device.torch_device)
            )
            .long()
            .item()
        )

    @property
    def gnbytes(self) -> int:
        """
        Returns the number of bytes consumed by the global ``DNDarray``

        Note
        -----------
            Does not include memory consumed by non-element attributes of the ``DNDarray`` object.
        """
        return self.nbytes

    @property
    def gnumel(self) -> int:
        """
        Returns the number of total elements of the ``DNDarray``
        """
        return self.size

    @property
    def imag(self) -> DNDarray:
        """
        Return the imaginary part of the ``DNDarray``.
        """
        return complex_math.imag(self)

    @property
    def lnbytes(self) -> int:
        """
        Returns the number of bytes consumed by the local ``torch.Tensor``

        Note
        -------------------
            Does not include memory consumed by non-element attributes of the ``DNDarray`` object.
        """
        return self.__array.element_size() * self.__array.nelement()

    @property
    def lnumel(self) -> int:
        """
        Number of elements of the ``DNDarray`` on each process
        """
        return np.prod(self.__array.shape)

    @property
    def lloc(self) -> Union[DNDarray, None]:
        """
        Local item setter and getter. i.e. this function operates on a local
        level and only on the PyTorch tensors composing the :class:`DNDarray`.
        This function uses the LocalIndex class. As getter, it returns a ``DNDarray``
        with the indices selected at a *local* level

        Parameters
        ----------
        key : int or slice or Tuple[int,...]
            Indices of the desired data.
        value : scalar, optional
            All types compatible with pytorch tensors, if none given then this is a getter function

        Examples
        --------
        >>> a = ht.zeros((4, 5), split=0)
        DNDarray([[0., 0., 0., 0., 0.],
                  [0., 0., 0., 0., 0.],
                  [0., 0., 0., 0., 0.],
                  [0., 0., 0., 0., 0.]], dtype=ht.float32, device=cpu:0, split=0)
        >>> a.lloc[1, 0:4]
        (1/2) tensor([0., 0., 0., 0.])
        (2/2) tensor([0., 0., 0., 0.])
        >>> a.lloc[1, 0:4] = torch.arange(1, 5)
        >>> a
        DNDarray([[0., 0., 0., 0., 0.],
                  [1., 2., 3., 4., 0.],
                  [0., 0., 0., 0., 0.],
                  [1., 2., 3., 4., 0.]], dtype=ht.float32, device=cpu:0, split=0)
        """
        return LocalIndex(self.__array)

    @property
    def lshape(self) -> Tuple[int]:
        """
        Returns the shape of the ``DNDarray`` on each node
        """
        return tuple(self.__array.shape)

    @property
    def lshape_map(self) -> torch.Tensor:
        """
        Returns the lshape map. If it hasn't been previously created then it will be created here.
        """
        return self.create_lshape_map()

    @property
    def real(self) -> DNDarray:
        """
        Return the real part of the ``DNDarray``.
        """
        return complex_math.real(self)

    @property
    def shape(self) -> Tuple[int]:
        """
        Returns the shape of the ``DNDarray`` as a whole
        """
        return self.__gshape

    @property
    def split(self) -> int:
        """
        Returns the axis on which the ``DNDarray`` is split
        """
        return self.__split

    @property
    def stride(self) -> Tuple[int]:
        """
        Returns the steps in each dimension when traversing a ``DNDarray``. torch-like usage: ``self.stride()``
        """
        return self.__array.stride

    @property
    def strides(self) -> Tuple[int]:
        """
        Returns bytes to step in each dimension when traversing a ``DNDarray``. numpy-like usage: ``self.strides()``
        """
        steps = list(self.larray.stride())
        itemsize = self.larray.storage().element_size()
        strides = tuple(step * itemsize for step in steps)
        return strides

    @property
    def T(self):
        """
        Reverse the dimensions of a DNDarray.
        """
        # specialty docs for this version of transpose. The transpose function is in heat/core/linalg/basics
        return linalg.transpose(self, axes=None)

    @property
    def array_with_halos(self) -> torch.Tensor:
        """
        Fetch halos of size ``halo_size`` from neighboring ranks and save them in ``self.halo_next``/``self.halo_prev``
        in case they are not already stored. If ``halo_size`` differs from the size of already stored halos,
        the are overwritten.
        """
        return self.__cat_halo()

    def __prephalo(self, start, end) -> torch.Tensor:
        """
        Extracts the halo indexed by start, end from ``self.array`` in the direction of ``self.split``

        Parameters
        ----------
        start : int
            Start index of the halo extracted from ``self.array``
        end : int
            End index of the halo extracted from ``self.array``
        """
        ix = [slice(None, None, None)] * len(self.shape)
        try:
            ix[self.split] = slice(start, end)
        except IndexError:
            print("Indices out of bound")

        return self.__array[ix].clone().contiguous()

    def get_halo(self, halo_size: int) -> torch.Tensor:
        """
        Fetch halos of size ``halo_size`` from neighboring ranks and save them in ``self.halo_next/self.halo_prev``.

        Parameters
        ----------
        halo_size : int
            Size of the halo.
        """
        if not isinstance(halo_size, int):
            raise TypeError(
                f"halo_size needs to be of Python type integer, {type(halo_size)} given"
            )
        if halo_size < 0:
            raise ValueError(
                f"halo_size needs to be a positive Python integer, {type(halo_size)} given"
            )

        if self.is_distributed() and halo_size > 0:
            # gather lshapes
            lshape_map = self.lshape_map
            rank = self.comm.rank

            populated_ranks = torch.nonzero(lshape_map[:, self.split]).squeeze().tolist()
            if rank in populated_ranks:
                first_rank = populated_ranks[0]
                last_rank = populated_ranks[-1]
                if rank != last_rank:
                    next_rank = populated_ranks[populated_ranks.index(rank) + 1]
                if rank != first_rank:
                    prev_rank = populated_ranks[populated_ranks.index(rank) - 1]
            else:
                # if process has no data we ignore it
                return

            if (halo_size > self.lshape_map[:, self.split][populated_ranks]).any():
                # halo_size is larger than the local size on at least one process
                raise ValueError(
                    f"halo_size {halo_size} needs to be smaller than chunk-size {self.lshape[self.split]} )"
                )

            a_prev = self.__prephalo(0, halo_size)
            a_next = self.__prephalo(-halo_size, None)
            res_prev = None
            res_next = None

            req_list = []

            # exchange data with next populated process
            if rank != last_rank:
                self.comm.Isend(a_next, next_rank)
                res_prev = torch.zeros(
                    a_prev.size(), dtype=a_prev.dtype, device=self.device.torch_device
                )
                req_list.append(self.comm.Irecv(res_prev, source=next_rank))

            if rank != first_rank:
                self.comm.Isend(a_prev, prev_rank)
                res_next = torch.zeros(
                    a_next.size(), dtype=a_next.dtype, device=self.device.torch_device
                )
                req_list.append(self.comm.Irecv(res_next, source=prev_rank))

            for req in req_list:
                req.Wait()

            self.__halo_next = res_prev
            self.__halo_prev = res_next
            self.__ishalo = True

    def __cat_halo(self) -> torch.Tensor:
        """
        Return local array concatenated to halos if they are available.
        """
        if not self.is_distributed():
            return self.__array
        return torch.cat(
            [_ for _ in (self.__halo_prev, self.__array, self.__halo_next) if _ is not None],
            dim=self.split,
        )

    def __array__(self) -> np.ndarray:
        """
        Returns a view of the process-local slice of the :class:`DNDarray` as a numpy ndarray, if the ``DNDarray`` resides on CPU. Otherwise, it returns a copy, on CPU, of the process-local slice of ``DNDarray`` as numpy ndarray.
        """
        return self.larray.cpu().__array__()

    def astype(self, dtype, copy=True) -> DNDarray:
        """
        Returns a casted version of this array.
        Casted array is a new array of the same shape but with given type of this array. If copy is ``True``, the
        same array is returned instead.

        Parameters
        ----------
        dtype : datatype
            Heat type to which the array is cast
        copy : bool, optional
            By default the operation returns a copy of this array. If copy is set to ``False`` the cast is performed
            in-place and this array is returned

        """
        dtype = canonical_heat_type(dtype)
        casted_array = self.__array.type(dtype.torch_type())
        if copy:
            return DNDarray(
                casted_array, self.shape, dtype, self.split, self.device, self.comm, self.balanced
            )

        self.__array = casted_array
        self.__dtype = dtype

        return self

    def balance_(self) -> DNDarray:
        """
        Function for balancing a :class:`DNDarray` between all nodes. To determine if this is needed use the :func:`is_balanced()` function.
        If the ``DNDarray`` is already balanced this function will do nothing. This function modifies the ``DNDarray``
        itself and will not return anything.

        Examples
        --------
        >>> a = ht.zeros((10, 2), split=0)
        >>> a[:, 0] = ht.arange(10)
        >>> b = a[3:]
        [0/2] tensor([[3., 0.],
        [1/2] tensor([[4., 0.],
                      [5., 0.],
                      [6., 0.]])
        [2/2] tensor([[7., 0.],
                      [8., 0.],
                      [9., 0.]])
        >>> b.balance_()
        >>> print(b.gshape, b.lshape)
        [0/2] (7, 2) (1, 2)
        [1/2] (7, 2) (3, 2)
        [2/2] (7, 2) (3, 2)
        >>> b
        [0/2] tensor([[3., 0.],
                     [4., 0.],
                     [5., 0.]])
        [1/2] tensor([[6., 0.],
                      [7., 0.]])
        [2/2] tensor([[8., 0.],
                      [9., 0.]])
        >>> print(b.gshape, b.lshape)
        [0/2] (7, 2) (3, 2)
        [1/2] (7, 2) (2, 2)
        [2/2] (7, 2) (2, 2)
        """
        if self.is_balanced(force_check=True):
            return
        self.redistribute_()

    def __bool__(self) -> bool:
        """
        Boolean scalar casting.
        """
        return self.__cast(bool)

    def __cast(self, cast_function) -> Union[float, int]:
        """
        Implements a generic cast function for ``DNDarray`` objects.

        Parameters
        ----------
        cast_function : function
            The actual cast function, e.g. ``float`` or ``int``

        Raises
        ------
        TypeError
            If the ``DNDarray`` object cannot be converted into a scalar.

        """
        if np.prod(self.shape) == 1:
            if self.split is None:
                return cast_function(self.__array)

            is_empty = np.prod(self.__array.shape) == 0
            root = self.comm.allreduce(0 if is_empty else self.comm.rank, op=MPI.SUM)

            return self.comm.bcast(None if is_empty else cast_function(self.__array), root=root)

        raise TypeError("only size-1 arrays can be converted to Python scalars")

    def __complex__(self) -> DNDarray:
        """
        Complex scalar casting.
        """
        return self.__cast(complex)

    def counts_displs(self) -> Tuple[Tuple[int], Tuple[int]]:
        """
        Returns actual counts (number of items per process) and displacements (offsets) of the DNDarray.
        Does not assume load balance.
        """
        if self.split is not None:
            counts = self.lshape_map[:, self.split]
            displs = [0] + torch.cumsum(counts, dim=0)[:-1].tolist()
            return tuple(counts.tolist()), tuple(displs)
        else:
            raise ValueError("Non-distributed DNDarray. Cannot calculate counts and displacements.")

    def cpu(self) -> DNDarray:
        """
        Returns a copy of this object in main memory. If this object is already in main memory, then no copy is
        performed and the original object is returned.
        """
        self.__array = self.__array.cpu()
        self.__device = devices.cpu
        return self

    def create_lshape_map(self, force_check: bool = False) -> torch.Tensor:
        """
        Generate a 'map' of the lshapes of the data on all processes.
        Units are ``(process rank, lshape)``

        Parameters
        ----------
        force_check : bool, optional
            if False (default) and the lshape map has already been created, use the previous
            result. Otherwise, create the lshape_map
        """
        if not force_check and self.__lshape_map is not None:
            return self.__lshape_map.clone()

        lshape_map = torch.zeros(
            (self.comm.size, self.ndim), dtype=torch.int64, device=self.device.torch_device
        )
        if not self.is_distributed:
            lshape_map[:] = torch.tensor(self.gshape, device=self.device.torch_device)
            return lshape_map
        if self.is_balanced(force_check=True):
            for i in range(self.comm.size):
                _, lshape, _ = self.comm.chunk(self.gshape, self.split, rank=i)
                lshape_map[i, :] = torch.tensor(lshape, device=self.device.torch_device)
        else:
            lshape_map[self.comm.rank, :] = torch.tensor(
                self.lshape, device=self.device.torch_device
            )
            self.comm.Allreduce(MPI.IN_PLACE, lshape_map, MPI.SUM)

        self.__lshape_map = lshape_map
        return lshape_map.clone()

    def create_partition_interface(self):
        """
        Create a partition interface in line with the DPPY proposal. This is subject to change.
        The intention of this to facilitate the usage of a general format for the referencing of
        distributed datasets.

        An example of the output and shape is shown below.

        __partitioned__ = {
            'shape': (27, 3, 2),
            'partition_tiling': (4, 1, 1),
            'partitions': {
                (0, 0, 0): {
                    'start': (0, 0, 0),
                    'shape': (7, 3, 2),
                    'data': tensor([...], dtype=torch.int32),
                    'location': [0],
                    'dtype': torch.int32,
                    'device': 'cpu'
                },
                (1, 0, 0): {
                    'start': (7, 0, 0),
                    'shape': (7, 3, 2),
                    'data': None,
                    'location': [1],
                    'dtype': torch.int32,
                    'device': 'cpu'
                },
                (2, 0, 0): {
                    'start': (14,  0,  0),
                    'shape': (7, 3, 2),
                    'data': None,
                    'location': [2],
                    'dtype': torch.int32,
                    'device': 'cpu'
                },
                (3, 0, 0): {
                    'start': (21,  0,  0),
                    'shape': (6, 3, 2),
                    'data': None,
                    'location': [3],
                    'dtype': torch.int32,
                    'device': 'cpu'
                }
            },
            'locals': [(rank, 0, 0)],
            'get': lambda x: x,
        }

        Returns
        -------
        dictionary containing the partition interface as shown above.
        """
        lshape_map = self.create_lshape_map()
        start_idx_map = torch.zeros_like(lshape_map)

        part_tiling = [1] * self.ndim
        lcls = [0] * self.ndim

        z = torch.tensor([0], device=self.device.torch_device, dtype=self.dtype.torch_type())
        if self.split is not None:
            starts = torch.cat((z, torch.cumsum(lshape_map[:, self.split], dim=0)[:-1]), dim=0)
            lcls[self.split] = self.comm.rank
            part_tiling[self.split] = self.comm.size
        else:
            starts = torch.zeros(self.ndim, dtype=torch.int, device=self.device.torch_device)

        start_idx_map[:, self.split] = starts

        partitions = {}
        base_key = [0] * self.ndim
        for r in range(self.comm.size):
            if self.split is not None:
                base_key[self.split] = r
                dat = None if r != self.comm.rank else self.larray
            else:
                dat = self.larray
            partitions[tuple(base_key)] = {
                "start": tuple(start_idx_map[r].tolist()),
                "shape": tuple(lshape_map[r].tolist()),
                "data": dat,
                "location": [r],
                "dtype": self.dtype.torch_type(),
                "device": self.device.torch_device,
            }

        partition_dict = {
            "shape": self.gshape,
            "partition_tiling": tuple(part_tiling),
            "partitions": partitions,
            "locals": [tuple(lcls)],
            "get": lambda x: x,
        }

        self.__partitions_dict__ = partition_dict

        return partition_dict

    def __float__(self) -> DNDarray:
        """
        Float scalar casting.

        See Also
        ---------
        :func:`~heat.core.manipulations.flatten`
        """
        return self.__cast(float)

    def fill_diagonal(self, value: float) -> DNDarray:
        """
        Fill the main diagonal of a 2D :class:`DNDarray`.
        This function modifies the input tensor in-place, and returns the input array.

        Parameters
        ----------
        value : float
            The value to be placed in the ``DNDarrays`` main diagonal
        """
        # Todo: make this 3D/nD
        if len(self.shape) != 2:
            raise ValueError("Only 2D tensors supported at the moment")

        if self.split is not None and self.comm.is_distributed:
            counts, displ, _ = self.comm.counts_displs_shape(self.shape, self.split)
            k = min(self.shape[0], self.shape[1])
            for p in range(self.comm.size):
                if displ[p] > k:
                    break
                proc = p
            if self.comm.rank <= proc:
                indices = (
                    displ[self.comm.rank],
                    displ[self.comm.rank + 1] if (self.comm.rank + 1) != self.comm.size else k,
                )
                if self.split == 0:
                    self.larray[:, indices[0] : indices[1]] = self.larray[
                        :, indices[0] : indices[1]
                    ].fill_diagonal_(value)
                elif self.split == 1:
                    self.larray[indices[0] : indices[1], :] = self.larray[
                        indices[0] : indices[1], :
                    ].fill_diagonal_(value)

        else:
            self.larray = self.larray.fill_diagonal_(value)

        return self

    def __getitem__(self, key: Union[int, Tuple[int, ...], List[int, ...]]) -> DNDarray:
        """
        Global getter function for DNDarrays.
        Returns a new DNDarray composed of the elements of the original tensor selected by the indices
        given. This does *NOT* redistribute or rebalance the resulting tensor. If the selection of values is
        unbalanced then the resultant tensor is also unbalanced!
        To redistributed the ``DNDarray`` use :func:`balance()` (issue #187)

        Parameters
        ----------
        key : int, slice, Tuple[int,...], List[int,...]
            Indices to get from the tensor.

        Examples
        --------
        >>> a = ht.arange(10, split=0)
        (1/2) >>> tensor([0, 1, 2, 3, 4], dtype=torch.int32)
        (2/2) >>> tensor([5, 6, 7, 8, 9], dtype=torch.int32)
        >>> a[1:6]
        (1/2) >>> tensor([1, 2, 3, 4], dtype=torch.int32)
        (2/2) >>> tensor([5], dtype=torch.int32)
        >>> a = ht.zeros((4,5), split=0)
        (1/2) >>> tensor([[0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0.]])
        (2/2) >>> tensor([[0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0.]])
        >>> a[1:4, 1]
        (1/2) >>> tensor([0.])
        (2/2) >>> tensor([0., 0.])
        """
        key = getattr(key, "copy()", key)
        l_dtype = self.dtype.torch_type()
        advanced_ind = False
        if isinstance(key, DNDarray) and key.ndim == self.ndim:
            """if the key is a DNDarray and it has as many dimensions as self, then each of the
            entries in the 0th dim refer to a single element. To handle this, the key is split
            into the torch tensors for each dimension. This signals that advanced indexing is
            to be used."""
            # NOTE: this gathers the entire key on every process!!
            # TODO: remove this resplit!!
            key = manipulations.resplit(key)
            if key.larray.dtype in [torch.bool, torch.uint8]:
                key = indexing.nonzero(key)

            if key.ndim > 1:
                key = list(key.larray.split(1, dim=1))
                # key is now a list of tensors with dimensions (key.ndim, 1)
                # squeeze singleton dimension:
                key = [key[i].squeeze_(1) for i in range(len(key))]
            else:
                key = [key]
            advanced_ind = True
        elif not isinstance(key, tuple):
            """this loop handles all other cases. DNDarrays which make it to here refer to
            advanced indexing slices, as do the torch tensors. Both DNDaarrys and torch.Tensors
            are cast into lists here by PyTorch. lists mean advanced indexing will be used"""
            h = [slice(None, None, None)] * max(self.ndim, 1)
            if isinstance(key, DNDarray):
                key = manipulations.resplit(key)
                if key.larray.dtype in [torch.bool, torch.uint8]:
                    h[0] = torch.nonzero(key.larray).flatten()  # .tolist()
                else:
                    h[0] = key.larray.tolist()
            elif isinstance(key, torch.Tensor):
                if key.dtype in [torch.bool, torch.uint8]:
                    # (coquelin77) i am not certain why this works without being a list. but it works...for now
                    h[0] = torch.nonzero(key).flatten()  # .tolist()
                else:
                    h[0] = key.tolist()
            else:
                h[0] = key

            key = list(h)

        if isinstance(key, (list, tuple)):
            key = list(key)
            for i, k in enumerate(key):
                # this might be a good place to check if the dtype is there
                try:
                    k = manipulations.resplit(k)
                    key[i] = k.larray
                except AttributeError:
                    pass

        # ellipsis
        key = list(key)
        key_classes = [type(n) for n in key]
        # if any(isinstance(n, ellipsis) for n in key):
        n_elips = key_classes.count(type(...))
        if n_elips > 1:
            raise ValueError("key can only contain 1 ellipsis")
        elif n_elips == 1:
            # get which item is the ellipsis
            ell_ind = key_classes.index(type(...))
            kst = key[:ell_ind]
            kend = key[ell_ind + 1 :]
            slices = [slice(None)] * (self.ndim - (len(kst) + len(kend)))
            key = kst + slices + kend
        else:
            key = key + [slice(None)] * (self.ndim - len(key))

        self_proxy = self.__torch_proxy__()
        for i in range(len(key)):
            if self.__key_adds_dimension(key, i, self_proxy):
                key[i] = slice(None)
                return self.expand_dims(i)[tuple(key)]

        key = tuple(key)
        # assess final global shape
        gout_full = list(self_proxy[key].shape)

        # calculate new split axis
        new_split = self.split
        # when slicing, squeezed singleton dimensions may affect new split axis
        if self.split is not None and len(gout_full) < self.ndim:
            if advanced_ind:
                new_split = 0
            else:
                for i in range(len(key[: self.split + 1])):
                    if self.__key_is_singular(key, i, self_proxy):
                        new_split = None if i == self.split else new_split - 1

        key = tuple(key)
        if not self.is_distributed():
            arr = self.__array[key].reshape(gout_full)
            return DNDarray(
                arr, tuple(gout_full), self.dtype, new_split, self.device, self.comm, self.balanced
            )

        # else: (DNDarray is distributed)
        arr = torch.tensor([], dtype=self.__array.dtype, device=self.__array.device)
        rank = self.comm.rank
        counts, chunk_starts = self.counts_displs()
        counts, chunk_starts = torch.tensor(counts), torch.tensor(chunk_starts)
        chunk_ends = chunk_starts + counts
        chunk_start = chunk_starts[rank]
        chunk_end = chunk_ends[rank]

        if len(key) == 0:  # handle empty list
            # this will return an array of shape (0, ...)
            arr = self.__array[key]

        """ At the end of the following if/elif/elif block the output array will be set.
            each block handles the case where the element of the key along the split axis
            is a different type and converts the key from global indices to local indices. """
        lout = gout_full.copy()

        if (
            isinstance(key[self.split], (list, torch.Tensor, DNDarray, np.ndarray))
            and len(key[self.split]) > 1
        ):
            # advanced indexing, elements in the split dimension are adjusted to the local indices
            lkey = list(key)
            if isinstance(key[self.split], DNDarray):
                lkey[self.split] = key[self.split].larray

            if not isinstance(lkey[self.split], torch.Tensor):
                inds = torch.tensor(
                    lkey[self.split], dtype=torch.long, device=self.device.torch_device
                )
            elif lkey[self.split].dtype in [torch.bool, torch.uint8]:  # or torch.byte?
                # need to convert the bools to indices
                inds = torch.nonzero(lkey[self.split])
            else:
                inds = lkey[self.split]
            # todo: remove where in favor of nonzero? might be a speed upgrade. testing required
            loc_inds = torch.where((inds >= chunk_start) & (inds < chunk_end))
            # if there are no local indices on a process, then `arr` is empty
            # if local indices exist:
            if len(loc_inds[0]) != 0:
                # select same local indices for other (non-split) dimensions if necessary
                for i, k in enumerate(lkey):
                    if isinstance(k, (list, torch.Tensor, DNDarray)) and i != self.split:
                        lkey[i] = k[loc_inds]
                # correct local indices for offset
                inds = inds[loc_inds] - chunk_start
                lkey[self.split] = inds
                lout[new_split] = len(inds)
                arr = self.__array[tuple(lkey)].reshape(tuple(lout))
            elif len(loc_inds[0]) == 0:
                if new_split is not None:
                    lout[new_split] = len(loc_inds[0])
                else:
                    lout = [0] * len(gout_full)
                arr = torch.tensor([], dtype=self.larray.dtype, device=self.larray.device).reshape(
                    tuple(lout)
                )

        elif isinstance(key[self.split], slice):
            # standard slicing along the split axis,
            # adjust the slice start, stop, and step, then run it on the processes which have the requested data
            key = list(key)
            key[self.split] = stride_tricks.sanitize_slice(key[self.split], self.gshape[self.split])
            key_start, key_stop, key_step = (
                key[self.split].start,
                key[self.split].stop,
                key[self.split].step,
            )
            og_key_start = key_start
            st_pr = torch.where(key_start < chunk_ends)[0]
            st_pr = st_pr[0] if len(st_pr) > 0 else self.comm.size
            sp_pr = torch.where(key_stop >= chunk_starts)[0]
            sp_pr = sp_pr[-1] if len(sp_pr) > 0 else 0
            actives = list(range(st_pr, sp_pr + 1))
            if rank in actives:
                key_start = 0 if rank != actives[0] else key_start - chunk_starts[rank]
                key_stop = counts[rank] if rank != actives[-1] else key_stop - chunk_starts[rank]
                key_start, key_stop = self.__xitem_get_key_start_stop(
                    rank, actives, key_start, key_stop, key_step, chunk_ends, og_key_start
                )
                key[self.split] = slice(key_start, key_stop, key_step)
                lout[new_split] = (
                    math.ceil((key_stop - key_start) / key_step)
                    if key_step is not None
                    else key_stop - key_start
                )
                arr = self.__array[tuple(key)].reshape(lout)
            else:
                lout[new_split] = 0
                arr = torch.empty(lout, dtype=self.__array.dtype, device=self.__array.device)

        elif self.__key_is_singular(key, self.split, self_proxy):
            # getting one item along split axis:
            key = list(key)
            if isinstance(key[self.split], list):
                key[self.split] = key[self.split].pop()
            elif isinstance(key[self.split], (torch.Tensor, DNDarray, np.ndarray)):
                key[self.split] = key[self.split].item()
            # translate negative index
            if key[self.split] < 0:
                key[self.split] += self.gshape[self.split]

            active_rank = torch.where(key[self.split] >= chunk_starts)[0][-1].item()
            # slice `self` on `active_rank`, allocate `arr` on all other ranks in preparation for Bcast
            if rank == active_rank:
                key[self.split] -= chunk_start.item()
                arr = self.__array[tuple(key)].reshape(tuple(lout))
            else:
                arr = torch.empty(tuple(lout), dtype=self.larray.dtype, device=self.larray.device)
            # broadcast result
            # TODO: Replace with `self.comm.Bcast(arr, root=active_rank)` after fixing #784
            arr = self.comm.bcast(arr, root=active_rank)
            if arr.device != self.larray.device:
                # todo: remove when unnecessary (also after #784)
                arr = arr.to(device=self.larray.device)

        return DNDarray(
            arr.type(l_dtype),
            gout_full if isinstance(gout_full, tuple) else tuple(gout_full),
            self.dtype,
            new_split,
            self.device,
            self.comm,
            balanced=True if new_split is None else None,
        )

    if torch.cuda.device_count() > 0:

        def gpu(self) -> DNDarray:
            """
            Returns a copy of this object in GPU memory. If this object is already in GPU memory, then no copy is
            performed and the original object is returned.

            """
            self.__array = self.__array.cuda(devices.gpu.torch_device)
            self.__device = devices.gpu
            return self

    def __int__(self) -> DNDarray:
        """
        Integer scalar casting.
        """
        return self.__cast(int)

    def is_balanced(self, force_check: bool = False) -> bool:
        """
        Determine if ``self`` is balanced evenly (or as evenly as possible) across all nodes
        distributed evenly (or as evenly as possible) across all processes.
        This is equivalent to returning ``self.balanced``. If no information
        is available (``self.balanced = None``), the balanced status will be
        assessed via collective communication.

        Parameters
        force_check : bool, optional
            If True, the balanced status of the ``DNDarray`` will be assessed via
            collective communication in any case.
        """
        if not force_check and self.balanced is not None:
            return self.balanced

        _, _, chk = self.comm.chunk(self.shape, self.split)
        test_lshape = tuple([x.stop - x.start for x in chk])
        balanced = 1 if test_lshape == self.lshape else 0

        out = self.comm.allreduce(balanced, MPI.SUM)
        balanced = True if out == self.comm.size else False
        return balanced

    def is_distributed(self) -> bool:
        """
        Determines whether the data of this ``DNDarray`` is distributed across multiple processes.
        """
        return self.split is not None and self.comm.is_distributed()

    @staticmethod
    def __key_is_singular(key: any, axis: int, self_proxy: torch.Tensor) -> bool:
        # determine if the key gets a singular item
        zeros = (0,) * (self_proxy.ndim - 1)
        return self_proxy[(*zeros[:axis], key[axis], *zeros[axis:])].ndim == 0

    @staticmethod
    def __key_adds_dimension(key: any, axis: int, self_proxy: torch.Tensor) -> bool:
        # determine if the key adds a new dimension
        zeros = (0,) * (self_proxy.ndim - 1)
        return self_proxy[(*zeros[:axis], key[axis], *zeros[axis:])].ndim == 2

    def item(self):
        """
        Returns the only element of a 1-element :class:`DNDarray`.
        Mirror of the pytorch command by the same name. If size of ``DNDarray`` is >1 element, then a ``ValueError`` is
        raised (by pytorch)

        Examples
        -------
        >>> import heat as ht
        >>> x = ht.zeros((1))
        >>> x.item()
        0.0
        """
        if self.size > 1:
            raise ValueError("only one-element DNDarrays can be converted to Python scalars")
        # make sure the element is on every process
        self.resplit_(None)
        return self.__array.item()

    def __len__(self) -> int:
        """
        The length of the ``DNDarray``, i.e. the number of items in the first dimension.
        """
        return self.shape[0]

    def numpy(self) -> np.array:
        """
        Returns a copy of the :class:`DNDarray` as numpy ndarray. If the ``DNDarray`` resides on the GPU, the underlying data will be copied to the CPU first.

        If the ``DNDarray`` is distributed, an MPI Allgather operation will be performed before converting to np.ndarray, i.e. each MPI process will end up holding a copy of the entire array in memory.  Make sure process memory is sufficient!

        Examples
        --------
        >>> import heat as ht
        T1 = ht.random.randn((10,8))
        T1.numpy()
        """
        dist = self.copy().resplit_(axis=None)
        return dist.larray.cpu().numpy()

    def __repr__(self) -> str:
        """
        Computes a printable representation of the passed DNDarray.
        """
        return printing.__str__(self)

    def ravel(self):
        """
        Flattens the ``DNDarray``.

        See Also
        --------
        :func:`~heat.core.manipulations.ravel`

        Examples
        --------
        >>> a = ht.ones((2,3), split=0)
        >>> b = a.ravel()
        >>> a[0,0] = 4
        >>> b
        DNDarray([4., 1., 1., 1., 1., 1.], dtype=ht.float32, device=cpu:0, split=0)
        """
        return manipulations.ravel(self)

    def redistribute_(
        self, lshape_map: Optional[torch.Tensor] = None, target_map: Optional[torch.Tensor] = None
    ):
        """
        Redistributes the data of the :class:`DNDarray` *along the split axis* to match the given target map.
        This function does not modify the non-split dimensions of the ``DNDarray``.
        This is an abstraction and extension of the balance function.

        Parameters
        ----------
        lshape_map : torch.Tensor, optional
            The current lshape of processes.
            Units are ``[rank, lshape]``.
        target_map : torch.Tensor, optional
            The desired distribution across the processes.
            Units are ``[rank, target lshape]``.
            Note: the only important parts of the target map are the values along the split axis,
            values which are not along this axis are there to mimic the shape of the ``lshape_map``.

        Examples
        --------
        >>> st = ht.ones((50, 81, 67), split=2)
        >>> target_map = torch.zeros((st.comm.size, 3), dtype=torch.int64)
        >>> target_map[0, 2] = 67
        >>> print(target_map)
        [0/2] tensor([[ 0,  0, 67],
        [0/2]         [ 0,  0,  0],
        [0/2]         [ 0,  0,  0]], dtype=torch.int32)
        [1/2] tensor([[ 0,  0, 67],
        [1/2]         [ 0,  0,  0],
        [1/2]         [ 0,  0,  0]], dtype=torch.int32)
        [2/2] tensor([[ 0,  0, 67],
        [2/2]         [ 0,  0,  0],
        [2/2]         [ 0,  0,  0]], dtype=torch.int32)
        >>> print(st.lshape)
        [0/2] (50, 81, 23)
        [1/2] (50, 81, 22)
        [2/2] (50, 81, 22)
        >>> st.redistribute_(target_map=target_map)
        >>> print(st.lshape)
        [0/2] (50, 81, 67)
        [1/2] (50, 81, 0)
        [2/2] (50, 81, 0)
        """
        if not self.is_distributed():
            return
        snd_dtype = self.dtype.torch_type()
        # units -> {pr, 1st index, 2nd index}
        if lshape_map is None:
            # NOTE: giving an lshape map which is incorrect will result in an incorrect distribution
            lshape_map = self.create_lshape_map(force_check=True)
        else:
            if not isinstance(lshape_map, torch.Tensor):
                raise TypeError(f"lshape_map must be a torch.Tensor, currently {type(lshape_map)}")
            if lshape_map.shape != (self.comm.size, len(self.gshape)):
                raise ValueError(
                    f"lshape_map must have the shape ({self.comm.size}, {len(self.gshape)}), currently {lshape_map.shape}"
                )
        if target_map is None:  # if no target map is given then it will balance the tensor
            _, _, chk = self.comm.chunk(self.shape, self.split)
            target_map = lshape_map.clone()
            target_map[..., self.split] = 0
            for pr in range(self.comm.size):
                target_map[pr, self.split] = self.comm.chunk(self.shape, self.split, rank=pr)[1][
                    self.split
                ]
            self.__balanced = True
        else:
            sanitation.sanitize_in_tensor(target_map)
            if target_map[..., self.split].sum() != self.shape[self.split]:
                raise ValueError(
                    f"Sum along the split axis of the target map must be equal to the shape in that dimension, currently {target_map[..., self.split]}"
                )
            if target_map.shape != (self.comm.size, len(self.gshape)):
                raise ValueError(
                    f"target_map must have the shape {(self.comm.size, len(self.gshape))}, currently {target_map.shape}"
                )
            # no info on balanced status
            self.__balanced = False
        lshape_cumsum = torch.cumsum(lshape_map[..., self.split], dim=0)
        chunk_cumsum = torch.cat(
            (
                torch.tensor([0], device=self.device.torch_device),
                torch.cumsum(target_map[..., self.split], dim=0),
            ),
            dim=0,
        )
        # need the data start as well for process 0
        for rcv_pr in range(self.comm.size - 1):
            st = chunk_cumsum[rcv_pr].item()
            sp = chunk_cumsum[rcv_pr + 1].item()
            # start pr should be the next process with data
            if lshape_map[rcv_pr, self.split] >= target_map[rcv_pr, self.split]:
                # if there is more data on the process than the start process than start == stop
                st_pr = rcv_pr
                sp_pr = rcv_pr
            else:
                # if there is less data on the process than need to get the data from the next data
                # with data
                # need processes > rcv_pr with lshape > 0
                st_pr = (
                    torch.nonzero(input=lshape_map[rcv_pr:, self.split] > 0, as_tuple=False)[
                        0
                    ].item()
                    + rcv_pr
                )
                hld = (
                    torch.nonzero(input=sp <= lshape_cumsum[rcv_pr:], as_tuple=False).flatten()
                    + rcv_pr
                )
                sp_pr = hld[0].item() if hld.numel() > 0 else self.comm.size

            # st_pr and sp_pr are the processes on which the data sits at the beginning
            # need to loop from st_pr to sp_pr + 1 and send the pr
            for snd_pr in range(st_pr, sp_pr + 1):
                if snd_pr == self.comm.size:
                    break
                data_required = abs(sp - st - lshape_map[rcv_pr, self.split].item())
                send_amt = (
                    data_required
                    if data_required <= lshape_map[snd_pr, self.split]
                    else lshape_map[snd_pr, self.split]
                )
                if (sp - st) <= lshape_map[rcv_pr, self.split].item() or snd_pr == rcv_pr:
                    send_amt = 0
                # send amount is the data still needed by recv if that is available on the snd
                if send_amt != 0:
                    self.__redistribute_shuffle(
                        snd_pr=snd_pr, send_amt=send_amt, rcv_pr=rcv_pr, snd_dtype=snd_dtype
                    )
                lshape_cumsum[snd_pr] -= send_amt
                lshape_cumsum[rcv_pr] += send_amt
                lshape_map[rcv_pr, self.split] += send_amt
                lshape_map[snd_pr, self.split] -= send_amt
            if lshape_map[rcv_pr, self.split] > target_map[rcv_pr, self.split]:
                # if there is any data left on the process then send it to the next one
                send_amt = lshape_map[rcv_pr, self.split] - target_map[rcv_pr, self.split]
                self.__redistribute_shuffle(
                    snd_pr=rcv_pr, send_amt=send_amt.item(), rcv_pr=rcv_pr + 1, snd_dtype=snd_dtype
                )
                lshape_cumsum[rcv_pr] -= send_amt
                lshape_cumsum[rcv_pr + 1] += send_amt
                lshape_map[rcv_pr, self.split] -= send_amt
                lshape_map[rcv_pr + 1, self.split] += send_amt

        if any(lshape_map[..., self.split] != target_map[..., self.split]):
            # sometimes need to call the redistribute once more,
            # (in the case that the second to last processes needs to get data from +1 and -1)
            self.redistribute_(lshape_map=lshape_map, target_map=target_map)

        self.__lshape_map = target_map

    def __redistribute_shuffle(
        self,
        snd_pr: Union[int, torch.Tensor],
        send_amt: Union[int, torch.Tensor],
        rcv_pr: Union[int, torch.Tensor],
        snd_dtype: torch.dtype,
    ):
        """
        Function to abstract the function used during redistribute for shuffling data between
        processes along the split axis

        Parameters
        ----------
        snd_pr : int or torch.Tensor
            Sending process
        send_amt : int or torch.Tensor
            Amount of data to be sent by the sending process
        rcv_pr : int or torch.Tensor
            Receiving process
        snd_dtype : torch.dtype
            Torch type of the data in question
        """
        rank = self.comm.rank
        send_slice = [slice(None)] * self.ndim
        keep_slice = [slice(None)] * self.ndim
        if rank == snd_pr:
            if snd_pr < rcv_pr:  # data passed to a higher rank (off the bottom)
                send_slice[self.split] = slice(
                    self.lshape[self.split] - send_amt, self.lshape[self.split]
                )
                keep_slice[self.split] = slice(0, self.lshape[self.split] - send_amt)
            if snd_pr > rcv_pr:  # data passed to a lower rank (off the top)
                send_slice[self.split] = slice(0, send_amt)
                keep_slice[self.split] = slice(send_amt, self.lshape[self.split])
            data = self.__array[send_slice].clone()
            self.comm.Send(data, dest=rcv_pr, tag=685)
            self.__array = self.__array[keep_slice]
        if rank == rcv_pr:
            shp = list(self.gshape)
            shp[self.split] = send_amt
            data = torch.zeros(shp, dtype=snd_dtype, device=self.device.torch_device)
            self.comm.Recv(data, source=snd_pr, tag=685)
            if snd_pr < rcv_pr:  # data passed from a lower rank (append to top)
                self.__array = torch.cat((data, self.__array), dim=self.split)
            if snd_pr > rcv_pr:  # data passed from a higher rank (append to bottom)
                self.__array = torch.cat((self.__array, data), dim=self.split)

    def resplit_(self, axis: int = None):
        """
        In-place option for resplitting a :class:`DNDarray`.

        Parameters
        ----------
        axis : int
            The new split axis, ``None`` denotes gathering, an int will set the new split axis

        Examples
        --------
        >>> a = ht.zeros((4, 5,), split=0)
        >>> a.lshape
        (0/2) (2, 5)
        (1/2) (2, 5)
        >>> ht.resplit_(a, None)
        >>> a.split
        None
        >>> a.lshape
        (0/2) (4, 5)
        (1/2) (4, 5)
        >>> a = ht.zeros((4, 5,), split=0)
        >>> a.lshape
        (0/2) (2, 5)
        (1/2) (2, 5)
        >>> ht.resplit_(a, 1)
        >>> a.split
        1
        >>> a.lshape
        (0/2) (4, 3)
        (1/2) (4, 2)
        """
        # sanitize the axis to check whether it is in range
        axis = sanitize_axis(self.shape, axis)

        # early out for unchanged content
        if self.comm.size == 1:
            self.__split = axis
            if axis is None:
                self.__partitions_dict__ = None
        if axis == self.split:
            return self

        self.__partitions_dict__ = None

        if axis is None:
            gathered = torch.empty(
                self.shape, dtype=self.dtype.torch_type(), device=self.device.torch_device
            )
            counts, displs = self.counts_displs()
            self.comm.Allgatherv(self.__array, (gathered, counts, displs), recv_axis=self.split)
            self.__array = gathered
            self.__split = axis
            self.__lshape_map = None
            return self
        # tensor needs be split/sliced locally
        if self.split is None:
            # new_arr = self
            _, _, slices = self.comm.chunk(self.shape, axis)
            temp = self.__array[slices]
            self.__array = torch.empty((1,), device=self.device.torch_device)
            # necessary to clear storage of local __array
            self.__array = temp.clone().detach()
            self.__split = axis
            self.__lshape_map = None
            return self

        tiles = tiling.SplitTiles(self)
        new_tile_locs = tiles.set_tile_locations(
            split=axis, tile_dims=tiles.tile_dimensions, arr=self
        )
        rank = self.comm.rank
        # receive the data with non-blocking, save which process its from
        rcv = {}
        for rpr in range(self.comm.size):
            # need to get where the tiles are on the new one first
            # rpr is the destination
            new_locs = torch.where(new_tile_locs == rpr)
            new_locs = torch.stack([new_locs[i] for i in range(self.ndim)], dim=1)
            for i in range(new_locs.shape[0]):
                key = tuple(new_locs[i].tolist())
                spr = tiles.tile_locations[key].item()
                to_send = tiles[key]
                if spr == rank and spr != rpr:
                    self.comm.Send(to_send.clone(), dest=rpr, tag=rank)
                    del to_send
                elif spr == rpr == rank:
                    rcv[key] = [None, to_send]
                elif rank == rpr:
                    sz = tiles.get_tile_size(key)
                    buf = torch.zeros(
                        sz, dtype=self.dtype.torch_type(), device=self.device.torch_device
                    )
                    w = self.comm.Irecv(buf=buf, source=spr, tag=spr)
                    rcv[key] = [w, buf]
        dims = list(range(self.ndim))
        del dims[axis]
        sorted_keys = sorted(rcv.keys())
        # todo: reduce the problem to 1D cats for each dimension, then work up
        sz = self.comm.size
        arrays = []
        for prs in range(int(len(sorted_keys) / sz)):
            lp_keys = sorted_keys[prs * sz : (prs + 1) * sz]
            lp_arr = None
            for k in lp_keys:
                if rcv[k][0] is not None:
                    rcv[k][0].Wait()
                if lp_arr is None:
                    lp_arr = rcv[k][1]
                else:
                    lp_arr = torch.cat((lp_arr, rcv[k][1]), dim=dims[-1])
                del rcv[k]
            if lp_arr is not None:
                arrays.append(lp_arr)
        del dims[-1]
        # for 4 prs and 4 dims, arrays is now 16 elements long,
        # next need to group the each 4 (sz) and cat in the next dim
        for d in reversed(dims):
            new_arrays = []
            for prs in range(int(len(arrays) / sz)):
                new_arrays.append(torch.cat(arrays[prs * sz : (prs + 1) * sz], dim=d))
            arrays = new_arrays
            del d
        if len(arrays) == 1:
            arrays = arrays[0]

        self.__array = arrays
        self.__split = axis
        self.__lshape_map = None
        return self

    def __setitem__(
        self,
        key: Union[int, Tuple[int, ...], List[int, ...]],
        value: Union[float, DNDarray, torch.Tensor],
    ):
        """
        Global item setter

        Parameters
        ----------
        key : Union[int, Tuple[int,...], List[int,...]]
            Index/indices to be set
        value: Union[float, DNDarray,torch.Tensor]
            Value to be set to the specified positions in the DNDarray (self)

        Notes
        -----
        If a ``DNDarray`` is given as the value to be set then the split axes are assumed to be equal.
        If they are not, PyTorch will raise an error when the values are attempted to be set on the local array

        Examples
        --------
        >>> a = ht.zeros((4,5), split=0)
        (1/2) >>> tensor([[0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0.]])
        (2/2) >>> tensor([[0., 0., 0., 0., 0.],
                          [0., 0., 0., 0., 0.]])
        >>> a[1:4, 1] = 1
        >>> a
        (1/2) >>> tensor([[0., 0., 0., 0., 0.],
                          [0., 1., 0., 0., 0.]])
        (2/2) >>> tensor([[0., 1., 0., 0., 0.],
                          [0., 1., 0., 0., 0.]])
        """
        key = getattr(key, "copy()", key)
        try:
            if value.split != self.split:
                val_split = int(value.split)
                sp = self.split
                warnings.warn(
                    f"\nvalue.split {val_split} not equal to this DNDarray's split:"
                    f" {sp}. this may cause errors or unwanted behavior",
                    category=RuntimeWarning,
                )
        except (AttributeError, TypeError):
            pass

        # NOTE: for whatever reason, there is an inplace op which interferes with the abstraction
        # of this next block of code. this is shared with __getitem__. I attempted to abstract it
        # in a standard way, but it was causing errors in the test suite. If someone else is
        # motived to do this they are welcome to, but i have no time right now
        # print(key)
        if isinstance(key, DNDarray) and key.ndim == self.ndim:
            """if the key is a DNDarray and it has as many dimensions as self, then each of the
            entries in the 0th dim refer to a single element. To handle this, the key is split
            into the torch tensors for each dimension. This signals that advanced indexing is
            to be used."""
            key = manipulations.resplit(key)
            if key.larray.dtype in [torch.bool, torch.uint8]:
                key = indexing.nonzero(key)

            if key.ndim > 1:
                key = list(key.larray.split(1, dim=1))
                # key is now a list of tensors with dimensions (key.ndim, 1)
                # squeeze singleton dimension:
                key = [key[i].squeeze_(1) for i in range(len(key))]
            else:
                key = [key]
        elif not isinstance(key, tuple):
            """this loop handles all other cases. DNDarrays which make it to here refer to
            advanced indexing slices, as do the torch tensors. Both DNDaarrys and torch.Tensors
            are cast into lists here by PyTorch. lists mean advanced indexing will be used"""
            h = [slice(None, None, None)] * self.ndim
            if isinstance(key, DNDarray):
                key = manipulations.resplit(key)
                if key.larray.dtype in [torch.bool, torch.uint8]:
                    h[0] = torch.nonzero(key.larray).flatten()  # .tolist()
                else:
                    h[0] = key.larray.tolist()
            elif isinstance(key, torch.Tensor):
                if key.dtype in [torch.bool, torch.uint8]:
                    # (coquelin77) im not sure why this works without being a list...but it does...for now
                    h[0] = torch.nonzero(key).flatten()  # .tolist()
                else:
                    h[0] = key.tolist()
            else:
                h[0] = key
            key = list(h)

        # key must be torch-proof
        if isinstance(key, (list, tuple)):
            key = list(key)
            for i, k in enumerate(key):
                try:  # extract torch tensor
                    k = manipulations.resplit(k)
                    key[i] = k.larray
                except AttributeError:
                    pass
                # remove bools from a torch tensor in favor of indexes
                try:
                    if key[i].dtype in [torch.bool, torch.uint8]:
                        key[i] = torch.nonzero(key[i]).flatten()
                except (AttributeError, TypeError):
                    pass

        key = list(key)

        # ellipsis stuff
        key_classes = [type(n) for n in key]
        # if any(isinstance(n, ellipsis) for n in key):
        n_elips = key_classes.count(type(...))
        if n_elips > 1:
            raise ValueError("key can only contain 1 ellipsis")
        elif n_elips == 1:
            # get which item is the ellipsis
            ell_ind = key_classes.index(type(...))
            kst = key[:ell_ind]
            kend = key[ell_ind + 1 :]
            slices = [slice(None)] * (self.ndim - (len(kst) + len(kend)))
            key = kst + slices + kend
        # ---------- end ellipsis stuff -------------

        for c, k in enumerate(key):
            try:
                key[c] = k.item()
            except (AttributeError, ValueError, RuntimeError):
                pass

        rank = self.comm.rank
        if self.split is not None:
            counts, chunk_starts = self.counts_displs()
        else:
            counts, chunk_starts = 0, [0] * self.comm.size
        counts = torch.tensor(counts, device=self.device.torch_device)
        chunk_starts = torch.tensor(chunk_starts, device=self.device.torch_device)
        chunk_ends = chunk_starts + counts
        chunk_start = chunk_starts[rank]
        chunk_end = chunk_ends[rank]
        # determine which elements are on the local process (if the key is a torch tensor)
        try:
            # if isinstance(key[self.split], torch.Tensor):
            filter_key = torch.nonzero(
                (chunk_start <= key[self.split]) & (key[self.split] < chunk_end)
            )
            for k in range(len(key)):
                try:
                    key[k] = key[k][filter_key].flatten()
                except TypeError:
                    pass
        except TypeError:  # this will happen if the key doesnt have that many
            pass

        key = tuple(key)

        if not self.is_distributed():
            return self.__setter(key, value)  # returns None

        # raise RuntimeError("split axis of array and the target value are not equal") removed
        # this will occur if the local shapes do not match
        rank = self.comm.rank
        ends = []
        for pr in range(self.comm.size):
            _, _, e = self.comm.chunk(self.shape, self.split, rank=pr)
            ends.append(e[self.split].stop - e[self.split].start)
        ends = torch.tensor(ends, device=self.device.torch_device)
        chunk_ends = ends.cumsum(dim=0)
        chunk_starts = torch.tensor([0] + chunk_ends.tolist(), device=self.device.torch_device)
        _, _, chunk_slice = self.comm.chunk(self.shape, self.split)
        chunk_start = chunk_slice[self.split].start
        chunk_end = chunk_slice[self.split].stop

        self_proxy = self.__torch_proxy__()

        # if the value is a DNDarray, the divisions need to be balanced:
        #   this means that we need to know how much data is where for both DNDarrays
        #   if the value data is not in the right place, then it will need to be moved

        if isinstance(key[self.split], slice):
            key = list(key)
            key_start = key[self.split].start if key[self.split].start is not None else 0
            key_stop = (
                key[self.split].stop
                if key[self.split].stop is not None
                else self.gshape[self.split]
            )
            if key_stop < 0:
                key_stop = self.gshape[self.split] + key[self.split].stop
            key_step = key[self.split].step
            og_key_start = key_start
            st_pr = torch.where(key_start < chunk_ends)[0]
            st_pr = st_pr[0] if len(st_pr) > 0 else self.comm.size
            sp_pr = torch.where(key_stop >= chunk_starts)[0]
            sp_pr = sp_pr[-1] if len(sp_pr) > 0 else 0
            actives = list(range(st_pr, sp_pr + 1))

            if (
                isinstance(value, type(self))
                and value.split is not None
                and value.shape[self.split] != self.shape[self.split]
            ):
                # setting elements in self with a DNDarray which is not the same size in the
                # split dimension
                local_keys = []
                # below is used if the target needs to be reshaped
                target_reshape_map = torch.zeros(
                    (self.comm.size, self.ndim), dtype=torch.int64, device=self.device.torch_device
                )
                for r in range(self.comm.size):
                    if r not in actives:
                        loc_key = key.copy()
                        loc_key[self.split] = slice(0, 0, 0)
                    else:
                        key_start_l = 0 if r != actives[0] else key_start - chunk_starts[r]
                        key_stop_l = ends[r] if r != actives[-1] else key_stop - chunk_starts[r]
                        key_start_l, key_stop_l = self.__xitem_get_key_start_stop(
                            r, actives, key_start_l, key_stop_l, key_step, chunk_ends, og_key_start
                        )
                        loc_key = key.copy()
                        loc_key[self.split] = slice(key_start_l, key_stop_l, key_step)

                        gout_full = torch.tensor(
                            self_proxy[loc_key].shape, device=self.device.torch_device
                        )
                        target_reshape_map[r] = gout_full
                    local_keys.append(loc_key)

                key = local_keys[rank]
                value = value.redistribute(target_map=target_reshape_map)

                if rank not in actives:
                    return  # non-active ranks can exit here

                chunk_starts_v = target_reshape_map[:, self.split]
                value_slice = [slice(None, None, None)] * value.ndim
                step2 = key_step if key_step is not None else 1
                key_start = (chunk_starts_v[rank] - og_key_start).item()

                key_start = max(key_start, 0)
                key_stop = key_start + key_stop
                slice_loc = min(self.split, value.ndim - 1)
                value_slice[slice_loc] = slice(
                    key_start, math.ceil(torch.true_divide(key_stop, step2)), 1
                )

                self.__setter(tuple(key), value.larray)
                return

            # if rank in actives:
            if rank not in actives:
                return  # non-active ranks can exit here
            key_start = 0 if rank != actives[0] else key_start - chunk_starts[rank]
            key_stop = ends[rank] if rank != actives[-1] else key_stop - chunk_starts[rank]
            key_start, key_stop = self.__xitem_get_key_start_stop(
                rank, actives, key_start, key_stop, key_step, chunk_ends, og_key_start
            )
            key[self.split] = slice(key_start, key_stop, key_step)

            # todo: need to slice the values to be the right size...
            if isinstance(value, (torch.Tensor, type(self))):
                # if its a torch tensor, it is assumed to exist on all processes
                value_slice = [slice(None, None, None)] * value.ndim
                step2 = key_step if key_step is not None else 1
                key_start = (chunk_starts[rank] - og_key_start).item()
                key_start = max(key_start, 0)
                key_stop = key_start + key_stop
                slice_loc = min(self.split, value.ndim - 1)
                value_slice[slice_loc] = slice(
                    key_start, math.ceil(torch.true_divide(key_stop, step2)), 1
                )
                self.__setter(tuple(key), value[tuple(value_slice)])
            else:
                self.__setter(tuple(key), value)
        elif isinstance(key[self.split], (torch.Tensor, list)):
            key = list(key)
            key[self.split] -= chunk_start
            if len(key[self.split]) != 0:
                self.__setter(tuple(key), value)

        elif key[self.split] in range(chunk_start, chunk_end):
            key = list(key)
            key[self.split] = key[self.split] - chunk_start
            self.__setter(tuple(key), value)

        elif key[self.split] < 0:
            key = list(key)
            if self.gshape[self.split] + key[self.split] in range(chunk_start, chunk_end):
                key[self.split] = key[self.split] + self.shape[self.split] - chunk_start
                self.__setter(tuple(key), value)

    def __setter(
        self,
        key: Union[int, Tuple[int, ...], List[int, ...]],
        value: Union[float, DNDarray, torch.Tensor],
    ):
        """
        Utility function for checking ``value`` and forwarding to :func:``__setitem__``

        Raises
        -------------
        NotImplementedError
            If the type of ``value`` ist not supported
        """
        if np.isscalar(value):
            self.__array.__setitem__(key, value)
        elif isinstance(value, DNDarray):
            self.__array.__setitem__(key, value.__array)
        elif isinstance(value, torch.Tensor):
            self.__array.__setitem__(key, value.data)
        elif isinstance(value, (list, tuple)):
            value = torch.tensor(value, device=self.device.torch_device)
            self.__array.__setitem__(key, value.data)
        elif isinstance(value, np.ndarray):
            value = torch.from_numpy(value)
            self.__array.__setitem__(key, value.data)
        else:
            raise NotImplementedError(f"Not implemented for {value.__class__.__name__}")

    def __str__(self) -> str:
        """
        Computes a string representation of the passed ``DNDarray``.
        """
        return printing.__str__(self)

    def tolist(self, keepsplit: bool = False) -> List:
        """
        Return a copy of the local array data as a (nested) Python list. For scalars, a standard Python number is returned.

        Parameters
        ----------
        keepsplit: bool
            Whether the list should be returned locally or globally.

        Examples
        --------
        >>> a = ht.array([[0,1],[2,3]])
        >>> a.tolist()
        [[0, 1], [2, 3]]

        >>> a = ht.array([[0,1],[2,3]], split=0)
        >>> a.tolist()
        [[0, 1], [2, 3]]

        >>> a = ht.array([[0,1],[2,3]], split=1)
        >>> a.tolist(keepsplit=True)
        (1/2) [[0], [2]]
        (2/2) [[1], [3]]
        """
        if not keepsplit:
            return self.resplit(axis=None).__array.tolist()

        return self.__array.tolist()

    def __torch_proxy__(self) -> torch.Tensor:
        """
        Return a 1-element `torch.Tensor` strided as the global `self` shape.
        Used internally for sanitation purposes.
        """
        return torch.ones((1,), dtype=torch.int8, device=self.larray.device).as_strided(
            self.gshape, [0] * self.ndim
        )

    @staticmethod
    def __xitem_get_key_start_stop(
        rank: int,
        actives: list,
        key_st: int,
        key_sp: int,
        step: int,
        ends: torch.Tensor,
        og_key_st: int,
    ) -> Tuple[int, int]:
        # this does some basic logic for adjusting the starting and stoping of the a key for
        #   setitem and getitem
        if step is not None and rank > actives[0]:
            offset = (ends[rank - 1] - og_key_st) % step
            if step > 2 and offset > 0:
                key_st += step - offset
            elif step == 2 and offset > 0:
                key_st += (ends[rank - 1] - og_key_st) % step
        if isinstance(key_st, torch.Tensor):
            key_st = key_st.item()
        if isinstance(key_sp, torch.Tensor):
            key_sp = key_sp.item()
        return key_st, key_sp


# HeAT imports at the end to break cyclic dependencies
from . import complex_math
from . import devices
from . import factories
from . import indexing
from . import linalg
from . import manipulations
from . import printing
from . import rounding
from . import sanitation
from . import statistics
from . import stride_tricks
from . import tiling

from .devices import Device
from .stride_tricks import sanitize_axis
from .types import datatype, canonical_heat_type



"""
This module handles operations focussing on complex numbers.
"""

import torch
from typing import Optional

from . import _operations
from . import constants
from . import factories
from . import trigonometrics
from . import types
from .dndarray import DNDarray

__all__ = ["angle", "conj", "conjugate", "imag", "real"]


def angle(x: DNDarray, deg: bool = False, out: Optional[DNDarray] = None) -> DNDarray:
    """
    Calculate the element-wise angle of the complex argument.

    Parameters
    ----------
    x : DNDarray
        Input array for which to compute the angle.
    deg : bool, optional
        Return the angle in degrees (True) or radiands (False).
    out : DNDarray, optional
        Output array with the angles.

    Examples
    --------
    >>> ht.angle(ht.array([1.0, 1.0j, 1+1j, -2+2j, 3 - 3j]))
    DNDarray([ 0.0000,  1.5708,  0.7854,  2.3562, -0.7854], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.angle(ht.array([1.0, 1.0j, 1+1j, -2+2j, 3 - 3j]), deg=True)
    DNDarray([  0.,  90.,  45., 135., -45.], dtype=ht.float32, device=cpu:0, split=None)
    """
    a = _operations.__local_op(torch.angle, x, out)

    if deg:
        a *= 180 / constants.pi

    return a


def conjugate(x: DNDarray, out: Optional[DNDarray] = None) -> DNDarray:
    """
    Compute the complex conjugate, element-wise.

    Parameters
    ----------
    x : DNDarray
        Input array for which to compute the complex conjugate.
    out : DNDarray, optional
        Output array with the complex conjugates.

    Examples
    --------
    >>> ht.conjugate(ht.array([1.0, 1.0j, 1+1j, -2+2j, 3 - 3j]))
    DNDarray([ (1-0j),     -1j,  (1-1j), (-2-2j),  (3+3j)], dtype=ht.complex64, device=cpu:0, split=None)
    """
    return _operations.__local_op(torch.conj, x, out)


# alias
conj = conjugate

# DNDarray method
DNDarray.conj = lambda self, out=None: conjugate(self, out)
DNDarray.conj.__doc__ = conjugate.__doc__


def imag(x: DNDarray) -> DNDarray:
    """
    Return the imaginary part of the complex argument. The returned DNDarray and the input DNDarray share the same underlying storage.

    Parameters
    ----------
    x : DNDarray
        Input array for which the imaginary part is returned.

    Examples
    --------
    >>> ht.imag(ht.array([1.0, 1.0j, 1+1j, -2+2j, 3 - 3j]))
    DNDarray([ 0.,  1.,  1.,  2., -3.], dtype=ht.float32, device=cpu:0, split=None)
    """
    if types.heat_type_is_complexfloating(x.dtype):
        return _operations.__local_op(torch.imag, x, None)
    else:
        return factories.zeros_like(x)


def real(x: DNDarray) -> DNDarray:
    """
    Return the real part of the complex argument. The returned DNDarray and the input DNDarray share the same underlying storage.

    Parameters
    ----------
    x : DNDarray
        Input array for which the real part is returned.

    Examples
    --------
    >>> ht.real(ht.array([1.0, 1.0j, 1+1j, -2+2j, 3 - 3j]))
    DNDarray([ 1.,  0.,  1., -2.,  3.], dtype=ht.float32, device=cpu:0, split=None)
    """
    if types.heat_type_is_complexfloating(x.dtype):
        return _operations.__local_op(torch.real, x, None)
    else:
        return x



"""This module contains Heat's version information."""


major: int = 1
"""Indicates Heat's main version."""
minor: int = 3
"""Indicates feature extension."""
micro: int = 1
"""Indicates revisions for bugfixes."""
extension: str = None
"""Indicates special builds, e.g. for specific hardware."""

if not extension:
    __version__: str = f"{major}.{minor}.{micro}"
    """The combined version string, consisting out of major, minor, micro and possibly extension."""
else:
    __version__: str = f"{major}.{minor}.{micro}-{extension}"



"""
Module for graph-based classes
"""
from __future__ import annotations

from typing import Callable
import torch
import heat as ht
from heat.core.dndarray import DNDarray


class Laplacian:
    """
    Graph Laplacian from a dataset

    Parameters
    ----------
    similarity : Callable
        Metric function that defines similarity between vertices. Should accept a data matrix :math:`n \\times f` as input and
        return an :math:`n\\times n` similarity matrix. Additional required parameters can be passed via a lambda function.
    definition : str
        Type of Laplacian \n
            - ``'simple'``: Laplacian matrix for simple graphs :math:`L = D - A` \n
            - ``'norm_sym'``: Symmetric normalized Laplacian :math:`L^{sym} = I - D^{-1/2} A D^{-1/2}` \n
            - ``'norm_rw'``: Random walk normalized Laplacian :math:`L^{rw} = D^{-1} L = I - D^{-1}` \n
    mode : str
        How to calculate adjacency from the similarity matrix \n
            - ``'fully_connected'`` is fully-connected, so :math:`A = S` \n
            - ``'eNeighbour'`` is the epsilon neighbourhood, with :math:`A_{ji} = 0` if :math:`S_{ij} > upper` or
            :math:`S_{ij} < lower`; for eNeighbour an upper or lower boundary needs to be set \n
    threshold_key : str
        ``'upper'`` or ``'lower'``, defining the type of threshold for the epsilon-neighborhood
    threshold_value : float
        Boundary value for the epsilon-neighborhood
    neighbours : int
        Number of nearest neighbors to be considered for adjacency definition. Currently not implemented
    """

    def __init__(
        self,
        similarity: Callable,
        weighted: bool = True,
        definition: str = "norm_sym",
        mode: str = "fully_connected",
        threshold_key: str = "upper",
        threshold_value: float = 1.0,
        neighbours: int = 10,
    ) -> DNDarray:
        self.similarity_metric = similarity
        self.weighted = weighted
        if definition not in ["simple", "norm_sym"]:
            raise NotImplementedError(
                "Currently only simple and normalized symmetric graph laplacians are supported"
            )
        else:
            self.definition = definition
        if mode not in ["eNeighbour", "fully_connected"]:
            raise NotImplementedError(
                "Only eNeighborhood and fully-connected graphs supported at the moment."
            )
        else:
            self.mode = mode

        if threshold_key not in ["upper", "lower"]:
            raise ValueError(
                "Only 'upper' and 'lower' threshold types supported for eNeighbouhood graph construction"
            )
        else:
            self.epsilon = (threshold_key, threshold_value)

        self.neighbours = neighbours

    def _normalized_symmetric_L(self, A: DNDarray) -> DNDarray:
        """
        Helper function to calculate the normalized symmetric Laplacian

        .. math:: L^{sym} = D^{-1/2} L D^{-1/2} = I - D^{-1/2} A D^{-1/2}

        Parameters
        ----------
        A : DNDarray
            The adjacency matrix of the graph
        """
        degree = ht.sum(A, axis=1)
        degree.resplit_(axis=None)
        # Find stand-alone vertices with no connections
        temp = torch.ones(
            degree.shape, dtype=degree.larray.dtype, device=degree.device.torch_device
        )
        degree.larray = torch.where(degree.larray == 0, temp, degree.larray)
        L = A / ht.sqrt(ht.expand_dims(degree, axis=1))
        L = L / ht.sqrt(ht.expand_dims(degree, axis=0))
        L = L * (-1.0)
        L.fill_diagonal(1.0)
        return L

    def _simple_L(self, A: DNDarray):
        """
        Helper function to calculate the simple graph Laplacian

        .. math:: L = D - A

        Parameters
        ----------
        A : DNDarray
            The Adjacency Matrix of the graph
        """
        degree = ht.sum(A, axis=1)
        L = ht.diag(degree) - A
        return L

    def construct(self, X: DNDarray) -> DNDarray:
        """
        Callable to get the Laplacian matrix from the dataset ``X`` according to the specified Laplacian

        Parameters
        ----------
        X : DNDarray
            The data matrix, Shape = (n_samples, n_features)
        """
        S = self.similarity_metric(X)
        S.fill_diagonal(0.0)

        if self.mode == "eNeighbour":
            if self.epsilon[0] == "upper":
                if self.weighted:
                    S = ht.where(S < self.epsilon[1], S, 0)
                else:
                    S = ht.int(S < self.epsilon[1])
            else:
                if self.weighted:
                    S = ht.where(S > self.epsilon[1], S, 0)
                else:
                    S = ht.int(S > self.epsilon[1])

        if self.definition == "simple":
            L = self._simple_L(S)
        elif self.definition == "norm_sym":
            L = self._normalized_symmetric_L(S)

        return L



"""
File for the MNIST dataset definition in heat
"""

import torch

from torchvision import datasets
from typing import Callable, Union

from ...core import factories
from . import datatools

__all__ = ["MNISTDataset"]


class MNISTDataset(datasets.MNIST):
    """
    Dataset wrapper for `torchvision.datasets.MNIST <https://pytorch.org/vision/stable/datasets.html#torchvision.datasets.MNIST>`_.
    This implements all of the required functions mentioned in :class:`heat.utils.data.Dataset`. The ``__getitem__`` and ``__len__`` functions are inherited from
    `torchvision.datasets.MNIST <https://pytorch.org/vision/stable/datasets.html#torchvision.datasets.MNIST>`_.

    Parameters
    ----------
    root : str
        Directory containing the MNIST dataset
    train : bool, optional
        If the data is the training dataset or not, default is True
    transform : Callable, optional
        Transform to be applied to the data dataset in the ``__getitem__`` function, default is ``None``
    target_transform : Callable, optional
        Transform to be applied to the target dataset in the ``__getitem__`` function, default is ``None``
    download : bool, optional
        If the data does not exist in the directory, download it if True (default)
    split : int, optional
        On which access to split the data when it is loaded into a ``DNDarray``
    ishuffle : bool, optional
        Flag indicating whether to use non-blocking communications for shuffling the data between epochs
        Note: if True, the ``Ishuffle()`` function must be defined within the class
        Default: ``False``
    test_set : bool, optional
        If this dataset is the testing set then keep all of the data local
        Default: ``False``

    Attributes
    ----------
    htdata : DNDarray
        full data
    httargets : DNDarray
        full target data
    comm : communication.MPICommunicator
        heat communicator for sending data between processes
    _cut_slice : slice
        slice to remove the last element if all are not equal in length
    lcl_half : int
        integer value of half of the data on the process
    data : torch.Tensor
        the local data on a process
    targets : torch.Tensor
        the local targets on a process
    ishuffle : bool
        flag indicating if non-blocking communications are used for shuffling the data between epochs
    test_set : bool
        if this dataset is the testing set then keep all of the data local

    Notes
    -----
    For other attributes see `torchvision.datasets.MNIST <https://pytorch.org/vision/stable/datasets.html#torchvision.datasets.MNIST>`_.
    """

    def __init__(
        self,
        root: str,
        train: bool = True,
        transform: Callable = None,
        target_transform: Callable = None,
        download: bool = True,
        split: int = 0,
        ishuffle: bool = False,
        test_set: bool = False,
    ):  # noqa: D107
        super().__init__(
            root,
            train=train,
            transform=transform,
            target_transform=target_transform,
            download=download,
        )
        if split != 0 and split is not None:
            raise ValueError("split must be 0 or None")
        split = None if test_set else split
        array = factories.array(self.data, split=split)
        targets = factories.array(self.targets, split=split)
        self.test_set = test_set
        self.partial_dataset = False
        self.comm = array.comm
        self.htdata = array
        self.httargets = targets
        self.ishuffle = ishuffle
        if split is not None:
            min_data_split = array.gshape[0] // array.comm.size
            arb_slice = slice(min_data_split)
            self._cut_slice = arb_slice
            self.lcl_half = min_data_split // 2
            self.data = array._DNDarray__array[self._cut_slice]
            self.targets = targets._DNDarray__array[self._cut_slice]
        else:
            self._cut_slice = None
            self.lcl_half = array.gshape[0] // 2
            self.data = array._DNDarray__array
            self.targets = targets._DNDarray__array
        # getitem and len are defined by torch's MNIST class

    def Shuffle(self):
        """
        Uses the :func:`datatools.dataset_shuffle` function to shuffle the data between the processes
        """
        if not self.test_set:
            datatools.dataset_shuffle(
                dataset=self, attrs=[["data", "htdata"], ["targets", "httargets"]]
            )

    def Ishuffle(self):
        """
        Uses the :func:`datatools.dataset_ishuffle` function to shuffle the data between the processes
        """
        if not self.test_set:
            datatools.dataset_ishuffle(
                dataset=self, attrs=[["data", "htdata"], ["targets", "httargets"]]
            )



"""
add modules/namespaces to the heat namespace
"""

from .core import *
from .core.linalg import *
from .core import __version__

from . import core
from . import classification
from . import cluster
from . import graph
from . import naive_bayes
from . import nn
from . import optim
from . import regression
from . import sparse
from . import spatial
from . import utils



"""
QR decomposition of (distributed) 2-D ``DNDarray``s.
"""
import collections
import torch
from typing import Type, Callable, Dict, Any, TypeVar, Union, Tuple

from ..communication import MPICommunication
from ..types import datatype
from ..tiling import SquareDiagTiles
from ..dndarray import DNDarray
from .. import factories

__all__ = ["qr"]


def qr(
    a: DNDarray,
    tiles_per_proc: Union[int, torch.Tensor] = 1,
    calc_q: bool = True,
    overwrite_a: bool = False,
) -> Tuple[DNDarray, DNDarray]:
    r"""
    Calculates the QR decomposition of a 2D ``DNDarray``.
    Factor the matrix ``a`` as *QR*, where ``Q`` is orthonormal and ``R`` is upper-triangular.
    If ``calc_q==True``, function returns ``QR(Q=Q, R=R)``, else function returns ``QR(Q=None, R=R)``

    Parameters
    ----------
    a : DNDarray
        Array which will be decomposed
    tiles_per_proc : int or torch.Tensor, optional
        Number of tiles per process to operate on,
    calc_q : bool, optional
        Whether or not to calculate Q.
        If ``True``, function returns ``(Q, R)``.
        If ``False``, function returns ``(None, R)``.
    overwrite_a : bool, optional
        If ``True``, function overwrites ``a`` with R
        If ``False``, a new array will be created for R

    Notes
    -----
    This function is built on top of PyTorch's QR function. ``torch.linalg.qr()`` using LAPACK on
    the backend.
    Basic information about QR factorization/decomposition can be found at
    https://en.wikipedia.org/wiki/QR_factorization. The algorithms are based on the CAQR and TSQRalgorithms. For more information see references.

    References
    ----------
    [0] W. Zheng, F. Song, L. Lin, and Z. Chen, “Scaling Up Parallel Computation of Tiled QR
    Factorizations by a Distributed Scheduling Runtime System and Analytical Modeling,”
    Parallel Processing Letters, vol. 28, no. 01, p. 1850004, 2018. \n
    [1] Bilel Hadri, Hatem Ltaief, Emmanuel Agullo, Jack Dongarra. Tile QR Factorization with
    Parallel Panel Processing for Multicore Architectures. 24th IEEE International Parallel
    and DistributedProcessing Symposium (IPDPS 2010), Apr 2010, Atlanta, United States.
    inria-00548899 \n
    [2] Gene H. Golub and Charles F. Van Loan. 1996. Matrix Computations (3rd Ed.).

    Examples
    --------
    >>> a = ht.random.randn(9, 6, split=0)
    >>> qr = ht.linalg.qr(a)
    >>> print(ht.allclose(a, ht.dot(qr.Q, qr.R)))
    [0/1] True
    [1/1] True
    >>> st = torch.randn(9, 6)
    >>> a = ht.array(st, split=1)
    >>> a_comp = ht.array(st, split=0)
    >>> q, r = ht.linalg.qr(a)
    >>> print(ht.allclose(a_comp, ht.dot(q, r)))
    [0/1] True
    [1/1] True
    """
    if not isinstance(a, DNDarray):
        raise TypeError("'a' must be a DNDarray")
    if not isinstance(tiles_per_proc, (int, torch.Tensor)):
        raise TypeError(
            f"tiles_per_proc must be an int or a torch.Tensor, currently {type(tiles_per_proc)}"
        )
    if not isinstance(calc_q, bool):
        raise TypeError(f"calc_q must be a bool, currently {type(calc_q)}")
    if not isinstance(overwrite_a, bool):
        raise TypeError(f"overwrite_a must be a bool, currently {type(overwrite_a)}")
    if isinstance(tiles_per_proc, torch.Tensor):
        raise ValueError(
            f"tiles_per_proc must be a single element torch.Tenor or int, currently has {tiles_per_proc.numel()} entries"
        )
    if len(a.shape) != 2:
        raise ValueError("Array 'a' must be 2 dimensional")

    QR = collections.namedtuple("QR", "Q, R")

    if a.split is None:
        try:
            q, r = torch.linalg.qr(a.larray, mode="complete")
        except AttributeError:
            q, r = a.larray.qr(some=False)

        q = factories.array(q, device=a.device, comm=a.comm)
        r = factories.array(r, device=a.device, comm=a.comm)
        ret = QR(q if calc_q else None, r)
        return ret
    # =============================== Prep work ====================================================
    r = a if overwrite_a else a.copy()
    # r.create_square_diag_tiles(tiles_per_proc=tiles_per_proc)
    r_tiles = SquareDiagTiles(arr=r, tiles_per_proc=tiles_per_proc)
    tile_columns = r_tiles.tile_columns
    tile_rows = r_tiles.tile_rows
    if calc_q:
        q = factories.eye(
            (r.gshape[0], r.gshape[0]), split=0, dtype=r.dtype, comm=r.comm, device=r.device
        )
        q_tiles = SquareDiagTiles(arr=q, tiles_per_proc=tiles_per_proc)
        q_tiles.match_tiles(r_tiles)
    else:
        q, q_tiles = None, None
    # ==============================================================================================

    if a.split == 0:
        rank = r.comm.rank
        active_procs = torch.arange(r.comm.size, device=r.device.torch_device)
        empties = torch.nonzero(input=r_tiles.lshape_map[..., 0] == 0, as_tuple=False)
        empties = empties[0] if empties.numel() > 0 else []
        for e in empties:
            active_procs = active_procs[active_procs != e]
        tile_rows_per_pr_trmd = r_tiles.tile_rows_per_process[: active_procs[-1] + 1]

        q_dict = {}
        q_dict_waits = {}
        proc_tile_start = torch.cumsum(
            torch.tensor(tile_rows_per_pr_trmd, device=r.device.torch_device), dim=0
        )
        # ------------------------------------ R Calculation ---------------------------------------
        for col in range(
            tile_columns
        ):  # for each tile column (need to do the last rank separately)
            # for each process need to do local qr
            not_completed_processes = torch.nonzero(
                input=col < proc_tile_start, as_tuple=False
            ).flatten()
            if rank not in not_completed_processes or rank not in active_procs:
                # if the process is done calculating R the break the loop
                break
            diag_process = not_completed_processes[0]
            __split0_r_calc(
                r_tiles=r_tiles,
                q_dict=q_dict,
                q_dict_waits=q_dict_waits,
                col_num=col,
                diag_pr=diag_process,
                not_completed_prs=not_completed_processes,
            )
        # ------------------------------------- Q Calculation --------------------------------------
        if calc_q:
            for col in range(tile_columns):
                __split0_q_loop(
                    col=col,
                    r_tiles=r_tiles,
                    proc_tile_start=proc_tile_start,
                    active_procs=active_procs,
                    q0_tiles=q_tiles,
                    q_dict=q_dict,
                    q_dict_waits=q_dict_waits,
                )
    elif a.split == 1:
        # loop over the tile columns
        lp_cols = tile_columns if a.gshape[0] > a.gshape[1] else tile_rows
        for dcol in range(lp_cols):  # dcol is the diagonal column
            __split1_qr_loop(dcol=dcol, r_tiles=r_tiles, q0_tiles=q_tiles, calc_q=calc_q)

    r.balance_()
    if q is not None:
        q.balance_()

    ret = QR(q, r)
    return ret


DNDarray.qr: Callable[
    [DNDarray, Union[int, torch.Tensor], bool, bool], Tuple[DNDarray, DNDarray]
] = lambda self, tiles_per_proc=1, calc_q=True, overwrite_a=False: qr(
    self, tiles_per_proc, calc_q, overwrite_a
)
DNDarray.qr.__doc__ = qr.__doc__


def __split0_global_q_dict_set(
    q_dict_col: Dict,
    col: Union[int, torch.Tensor],
    r_tiles: SquareDiagTiles,
    q_tiles: SquareDiagTiles,
    global_merge_dict: Dict = None,
) -> None:
    """
    The function takes the original Q tensors from the global QR calculation and sets them to
    the keys which corresponds with their tile coordinates in Q. this returns a separate dictionary,
    it does NOT set the values of Q

    Parameters
    ----------
    q_dict_col : Dict
        The dictionary of the Q values for a given column, should be given as q_dict[col]
    col : int or torch.Tensor
        current column for which Q is being calculated for
    r_tiles : SquareDiagTiles
        tiling object for ``r``
    q_tiles : SquareDiagTiles
        tiling object for Q0
    global_merge_dict : Dict, optional
        the output of the function will be in this dictionary
        Form of output: key index : ``torch.Tensor``

    """
    # q is already created, the job of this function is to create the group the merging q's together
    # it takes the merge qs, splits them, then puts them into a new dictionary
    proc_tile_start = torch.cumsum(
        torch.tensor(r_tiles.tile_rows_per_process, device=r_tiles.arr.larray.device), dim=0
    )
    diag_proc = torch.nonzero(input=proc_tile_start > col, as_tuple=False)[0].item()
    proc_tile_start = torch.cat(
        (torch.tensor([0], device=r_tiles.arr.larray.device), proc_tile_start[:-1]), dim=0
    )

    # 1: create caqr dictionary
    # need to have empty lists for all tiles in q
    global_merge_dict = {} if global_merge_dict is None else global_merge_dict

    # intended to be used as [row][column] -> data
    # 2: loop over keys in the dictionary
    merge_list = list(q_dict_col.keys())
    merge_list.sort()
    # todo: possible improvement -> make the keys have the process they are on as well,
    #  then can async get them if they are not on the diagonal process
    for key in merge_list:
        # this loops over all of the Qs for col and creates the dictionary for the pr Q merges
        p0 = key.find("p0")
        p1 = key.find("p1")
        end = key.find("e")
        r0 = int(key[p0 + 2 : p1])
        r1 = int(key[p1 + 2 : end])
        lp_q = q_dict_col[key][0]
        base_size = q_dict_col[key][1]
        # cut the q into 4 bits (end of base array)
        # todo: modify this so that it will get what is needed from the process,
        #  instead of gathering all the qs
        top_left = lp_q[: base_size[0], : base_size[0]]
        top_right = lp_q[: base_size[0], base_size[0] :]
        bottom_left = lp_q[base_size[0] :, : base_size[0]]
        bottom_right = lp_q[base_size[0] :, base_size[0] :]
        # need to adjust the keys to be the global row
        col1 = col if diag_proc == r0 else proc_tile_start[r0].item()
        col2 = proc_tile_start[r1].item()
        # col0 and col1 are the columns numbers
        # r0 and r1 are the ranks
        jdim = (col1, col1)
        kdim = (col1, col2)
        ldim = (col2, col1)
        mdim = (col2, col2)

        # if there are no elements on that location than set it as the tile
        # 1. get keys of what already has data
        curr_keys = set(global_merge_dict.keys())
        # 2. determine which tiles need to be touched/created
        # these are the keys which are to be multiplied by the q in the current loop
        # for matrix of form: | J  K |
        #                     | L  M |
        mult_keys_00 = [(i, col1) for i in range(q_tiles.tile_columns)]  # (J)
        # (J) -> inds: (i, col0)(col0, col0) -> set at (i, col0)
        mult_keys_01 = [(i, col1) for i in range(q_tiles.tile_columns)]  # (K)
        # (K) -> inds: (i, col0)(col0, col1) -> set at (i, col1)
        mult_keys_10 = [(i, col2) for i in range(q_tiles.tile_columns)]  # (L)
        # (L) -> inds: (i, col1)(col1, col0) -> set at (i, col0)
        mult_keys_11 = [(i, col2) for i in range(q_tiles.tile_columns)]  # (M)
        # (M) -> inds: (i, col1)(col1, col1) -> set at (i, col1)

        # if there are no elements in the mult_keys then set the element to the same place
        s00 = set(mult_keys_00) & curr_keys
        s01 = set(mult_keys_01) & curr_keys
        s10 = set(mult_keys_10) & curr_keys
        s11 = set(mult_keys_11) & curr_keys
        hold_dict = global_merge_dict.copy()

        # (J)
        if not len(s00):
            global_merge_dict[jdim] = top_left
        else:  # -> do the mm for all of the mult keys
            for k in s00:
                global_merge_dict[k[0], jdim[1]] = hold_dict[k] @ top_left
        # (K)
        if not len(s01):
            # check that we are not overwriting here
            global_merge_dict[kdim] = top_right
        else:  # -> do the mm for all of the mult keys
            for k in s01:
                global_merge_dict[k[0], kdim[1]] = hold_dict[k] @ top_right
        # (L)
        if not len(s10):
            # check that we are not overwriting here
            global_merge_dict[ldim] = bottom_left
        else:  # -> do the mm for all of the mult keys
            for k in s10:
                global_merge_dict[k[0], ldim[1]] = hold_dict[k] @ bottom_left
        # (M)
        if not len(s11):
            # check that we are not overwriting here
            global_merge_dict[mdim] = bottom_right
        else:  # -> do the mm for all of the mult keys
            for k in s11:
                global_merge_dict[k[0], mdim[1]] = hold_dict[k] @ bottom_right
    return global_merge_dict


def __split0_r_calc(
    r_tiles: SquareDiagTiles,
    q_dict: Dict,
    q_dict_waits: Dict,
    col_num: int,
    diag_pr: int,
    not_completed_prs: torch.Tensor,
) -> None:
    """
    Function to do the QR calculations to calculate the global R of the array ``a``.
    This function uses a binary merge structure in the global R merge.

    Parameters
    ----------
    r_tiles : SquareDiagTiles
        Tiling object for ``r``
    q_dict : Dict
        Dictionary to save the calculated Q matrices to
    q_dict_waits : Dict
        Dictionary to save the calculated Q matrices to which are
        not calculated on the diagonal process
    col_num : int
        The current column of the the R calculation
    diag_pr : int
        Rank of the process which has the tile which lies along the diagonal
    not_completed_prs : torch.Tensor
        Tensor of the processes which have not yet finished calculating R

    """
    tile_rows_proc = r_tiles.tile_rows_per_process
    comm = r_tiles.arr.comm
    rank = comm.rank
    lcl_tile_row = 0 if rank != diag_pr else col_num - sum(tile_rows_proc[:rank])
    # only work on the processes which have not computed the final result
    q_dict[col_num] = {}
    q_dict_waits[col_num] = {}

    # --------------- local QR calc -----------------------------------------------------
    base_tile = r_tiles.local_get(key=(slice(lcl_tile_row, None), col_num))
    try:
        q1, r1 = torch.linalg.qr(base_tile, mode="complete")
    except AttributeError:
        q1, r1 = base_tile.qr(some=False)

    q_dict[col_num]["l0"] = [q1, base_tile.shape]
    r_tiles.local_set(key=(slice(lcl_tile_row, None), col_num), value=r1)
    if col_num != r_tiles.tile_columns - 1:
        base_rest = r_tiles.local_get((slice(lcl_tile_row, None), slice(col_num + 1, None)))
        loc_rest = torch.matmul(q1.T, base_rest)
        r_tiles.local_set(key=(slice(lcl_tile_row, None), slice(col_num + 1, None)), value=loc_rest)
    # --------------- global QR calc (binary merge) -------------------------------------
    rem1 = None
    rem2 = None
    offset = not_completed_prs[0]
    loop_size_remaining = not_completed_prs.clone()
    completed = bool(loop_size_remaining.size()[0] <= 1)
    procs_remaining = loop_size_remaining.size()[0]
    loop = 0
    while not completed:
        if procs_remaining % 2 == 1:
            # if the number of processes active is odd need to save the remainders
            if rem1 is None:
                rem1 = loop_size_remaining[-1]
                loop_size_remaining = loop_size_remaining[:-1]
            elif rem2 is None:
                rem2 = loop_size_remaining[-1]
                loop_size_remaining = loop_size_remaining[:-1]
        if rank not in loop_size_remaining and rank not in [rem1, rem2]:
            break  # if the rank is done then exit the loop
        # send the data to the corresponding processes
        half_prs_rem = torch.div(procs_remaining, 2, rounding_mode="floor")

        zipped = zip(
            loop_size_remaining.flatten()[:half_prs_rem],
            loop_size_remaining.flatten()[half_prs_rem:],
        )
        for pr in zipped:
            pr0, pr1 = int(pr[0].item()), int(pr[1].item())
            __split0_merge_tile_rows(
                pr0=pr0,
                pr1=pr1,
                column=col_num,
                rank=rank,
                r_tiles=r_tiles,
                diag_process=diag_pr,
                key=str(loop) + "p0" + str(pr0) + "p1" + str(pr1) + "e",
                q_dict=q_dict,
            )

            __split0_send_q_to_diag_pr(
                col=col_num,
                pr0=pr0,
                pr1=pr1,
                diag_process=diag_pr,
                comm=comm,
                q_dict=q_dict,
                key=str(loop) + "p0" + str(pr0) + "p1" + str(pr1) + "e",
                q_dict_waits=q_dict_waits,
                q_dtype=r_tiles.arr.dtype.torch_type(),
                q_device=r_tiles.arr.larray.device,
            )

        loop_size_remaining = loop_size_remaining[: -1 * (half_prs_rem)]
        procs_remaining = loop_size_remaining.size()[0]

        if rem1 is not None and rem2 is not None:
            # combine rem1 and rem2 in the same way as the other nodes,
            # then save the results in rem1 to be used later
            __split0_merge_tile_rows(
                pr0=rem2,
                pr1=rem1,
                column=col_num,
                rank=rank,
                r_tiles=r_tiles,
                diag_process=diag_pr,
                key=str(loop) + "p0" + str(int(rem1)) + "p1" + str(int(rem2)) + "e",
                q_dict=q_dict if q_dict is not None else {},
            )

            rem1, rem2 = int(rem1), int(rem2)
            __split0_send_q_to_diag_pr(
                col=col_num,
                pr0=rem2,
                pr1=rem1,
                diag_process=diag_pr,
                key=str(loop) + "p0" + str(int(rem1)) + "p1" + str(int(rem2)) + "e",
                q_dict=q_dict if q_dict is not None else {},
                comm=comm,
                q_dict_waits=q_dict_waits,
                q_dtype=r_tiles.arr.dtype.torch_type(),
                q_device=r_tiles.arr.larray.device,
            )
            rem1 = rem2
            rem2 = None

        loop += 1
        if rem1 is not None and rem2 is None and procs_remaining == 1:
            # combine rem1 with process 0 (offset) and set completed to True
            # this should be the last thing that happens
            __split0_merge_tile_rows(
                pr0=offset,
                pr1=rem1,
                column=col_num,
                rank=rank,
                r_tiles=r_tiles,
                diag_process=diag_pr,
                key=str(loop) + "p0" + str(int(offset)) + "p1" + str(int(rem1)) + "e",
                q_dict=q_dict,
            )

            offset, rem1 = int(offset), int(rem1)
            __split0_send_q_to_diag_pr(
                col=col_num,
                pr0=offset,
                pr1=rem1,
                diag_process=diag_pr,
                key=str(loop) + "p0" + str(int(offset)) + "p1" + str(int(rem1)) + "e",
                q_dict=q_dict,
                comm=comm,
                q_dict_waits=q_dict_waits,
                q_dtype=r_tiles.arr.dtype.torch_type(),
                q_device=r_tiles.arr.larray.device,
            )
            rem1 = None

        completed = True if procs_remaining == 1 and rem1 is None and rem2 is None else False


def __split0_merge_tile_rows(
    pr0: int,
    pr1: int,
    column: int,
    rank: int,
    r_tiles: SquareDiagTiles,
    diag_process: int,
    key: str,
    q_dict: Dict,
) -> None:
    """
    Sets the value of ``q_dict[column][key]`` with ``[Q, upper.shape, lower.shape]``
    Merge two tile rows, take their QR, and apply it to the trailing process
    This will modify ``a`` and set the value of the ``q_dict[column][key]``
    with ``[Q, upper.shape, lower.shape]``.

    Parameters
    ----------
    pr0, pr1 : int, int
        Process ranks of the processes to be used
    column : int
        The current process of the QR calculation
    rank : int
        The rank of the process
    r_tiles : SquareDiagTiles
        Tiling object used for getting/setting the tiles required
    diag_process : int
        The rank of the process which has the tile along the diagonal for the given column
    key : str
        Input key
    q_dict : Dict
        Input dictionary
    """
    if rank not in [pr0, pr1]:
        return
    pr0 = pr0.item() if isinstance(pr0, torch.Tensor) else pr0
    pr1 = pr1.item() if isinstance(pr1, torch.Tensor) else pr1
    comm = r_tiles.arr.comm
    upper_row = sum(r_tiles.tile_rows_per_process[:pr0]) if pr0 != diag_process else column
    lower_row = sum(r_tiles.tile_rows_per_process[:pr1]) if pr1 != diag_process else column

    upper_inds = r_tiles.get_start_stop(key=(upper_row, column))
    lower_inds = r_tiles.get_start_stop(key=(lower_row, column))

    upper_size = (upper_inds[1] - upper_inds[0], upper_inds[3] - upper_inds[2])
    lower_size = (lower_inds[1] - lower_inds[0], lower_inds[3] - lower_inds[2])

    a_torch_device = r_tiles.arr.larray.device

    # upper adjustments
    if upper_size[0] < upper_size[1] and r_tiles.tile_rows_per_process[pr0] > 1:
        # end of dim0 (upper_inds[1]) is equal to the size in dim1
        upper_inds = list(upper_inds)
        upper_inds[1] = upper_inds[0] + upper_size[1]
        upper_size = (upper_inds[1] - upper_inds[0], upper_inds[3] - upper_inds[2])
    if lower_size[0] < lower_size[1] and r_tiles.tile_rows_per_process[pr1] > 1:
        # end of dim0 (upper_inds[1]) is equal to the size in dim1
        lower_inds = list(lower_inds)
        lower_inds[1] = lower_inds[0] + lower_size[1]
        lower_size = (lower_inds[1] - lower_inds[0], lower_inds[3] - lower_inds[2])

    if rank == pr0:
        # need to use lloc on r_tiles.arr with the indices
        upper = r_tiles.arr.lloc[upper_inds[0] : upper_inds[1], upper_inds[2] : upper_inds[3]]

        comm.Send(upper.clone(), dest=pr1, tag=986)
        lower = torch.zeros(lower_size, dtype=r_tiles.arr.dtype.torch_type(), device=a_torch_device)
        comm.Recv(lower, source=pr1, tag=4363)
    else:  # rank == pr1:
        lower = r_tiles.arr.lloc[lower_inds[0] : lower_inds[1], lower_inds[2] : lower_inds[3]]
        upper = torch.zeros(upper_size, dtype=r_tiles.arr.dtype.torch_type(), device=a_torch_device)
        comm.Recv(upper, source=pr0, tag=986)
        comm.Send(lower.clone(), dest=pr0, tag=4363)

    try:
        q_merge, r = torch.linalg.qr(torch.cat((upper, lower), dim=0), mode="complete")
    except AttributeError:
        q_merge, r = torch.cat((upper, lower), dim=0).qr(some=False)

    upp = r[: upper.shape[0]]
    low = r[upper.shape[0] :]
    if rank == pr0:
        r_tiles.arr.lloc[upper_inds[0] : upper_inds[1], upper_inds[2] : upper_inds[3]] = upp
    else:  # rank == pr1:
        r_tiles.arr.lloc[lower_inds[0] : lower_inds[1], lower_inds[2] : lower_inds[3]] = low

    if column < r_tiles.tile_columns - 1:
        upper_rest_size = (upper_size[0], r_tiles.arr.gshape[1] - upper_inds[3])
        lower_rest_size = (lower_size[0], r_tiles.arr.gshape[1] - lower_inds[3])

        if rank == pr0:
            upper_rest = r_tiles.arr.lloc[upper_inds[0] : upper_inds[1], upper_inds[3] :]
            lower_rest = torch.zeros(
                lower_rest_size, dtype=r_tiles.arr.dtype.torch_type(), device=a_torch_device
            )
            comm.Send(upper_rest.clone(), dest=pr1, tag=98654)
            comm.Recv(lower_rest, source=pr1, tag=436364)
        else:  # rank == pr1:
            lower_rest = r_tiles.arr.lloc[lower_inds[0] : lower_inds[1], lower_inds[3] :]
            upper_rest = torch.zeros(
                upper_rest_size, dtype=r_tiles.arr.dtype.torch_type(), device=a_torch_device
            )
            comm.Recv(upper_rest, source=pr0, tag=98654)
            comm.Send(lower_rest.clone(), dest=pr0, tag=436364)

        cat_tensor = torch.cat((upper_rest, lower_rest), dim=0)
        new_rest = torch.matmul(q_merge.t(), cat_tensor)
        # the data for upper rest is a slice of the new_rest, need to slice only the 0th dim
        upp = new_rest[: upper_rest.shape[0]]
        low = new_rest[upper_rest.shape[0] :]
        if rank == pr0:
            r_tiles.arr.lloc[upper_inds[0] : upper_inds[1], upper_inds[3] :] = upp
        # set the lower rest
        else:  # rank == pr1:
            r_tiles.arr.lloc[lower_inds[0] : lower_inds[1], lower_inds[3] :] = low

    q_dict[column][key] = [q_merge, upper.shape, lower.shape]


def __split0_send_q_to_diag_pr(
    col: int,
    pr0: int,
    pr1: int,
    diag_process: int,
    comm: MPICommunication,
    q_dict: Dict,
    key: str,
    q_dict_waits: Dict,
    q_dtype: Type[datatype],
    q_device: torch.device,
) -> None:
    """
    Sets the values of ``q_dict_waits`` with the with *waits* for the values of Q, ``upper.shape``,
    and ``lower.shape``
    This function sends the merged Q to the diagonal process. Buffered send it used for sending
    Q. This is needed for the Q calculation when two processes are merged and neither is the diagonal
    process.

    Parameters
    ----------
    col : int
        The current column used in the parent QR loop
    pr0, pr1 : int, int
        Rank of processes 0 and 1. These are the processes used in the calculation of q
    diag_process : int
        The rank of the process which has the tile along the diagonal for the given column
    comm : MPICommunication (ht.DNDarray.comm)
        The communicator used. (Intended as the communication of ``a`` given to qr)
    q_dict : Dict
        Dictionary containing the Q values calculated for finding R
    key : str
        Key for ``q_dict[col]`` which corresponds to the Q to send
    q_dict_waits : Dict
        Dictionary used in the collection of the Qs which are sent to the diagonal process
    q_dtype : torch.type
        Type of the Q tensor
    q_device : torch.device
        Torch device of the Q tensor

    """
    if comm.rank not in [pr0, pr1, diag_process]:
        return
    # this is to send the merged q to the diagonal process for the forming of q
    base_tag = "1" + str(pr1.item() if isinstance(pr1, torch.Tensor) else pr1)
    if comm.rank == pr1:
        q = q_dict[col][key][0]
        u_shape = q_dict[col][key][1]
        l_shape = q_dict[col][key][2]
        comm.send(tuple(q.shape), dest=diag_process, tag=int(base_tag + "1"))
        comm.Isend(q, dest=diag_process, tag=int(base_tag + "12"))
        comm.send(u_shape, dest=diag_process, tag=int(base_tag + "123"))
        comm.send(l_shape, dest=diag_process, tag=int(base_tag + "1234"))
    if comm.rank == diag_process:
        # q_dict_waits now looks like a
        q_sh = comm.recv(source=pr1, tag=int(base_tag + "1"))
        q_recv = torch.zeros(q_sh, dtype=q_dtype, device=q_device)
        k = "p0" + str(pr0) + "p1" + str(pr1)
        q_dict_waits[col][k] = []
        q_wait = comm.Irecv(q_recv, source=pr1, tag=int(base_tag + "12"))
        q_dict_waits[col][k].append([q_recv, q_wait])
        q_dict_waits[col][k].append(comm.irecv(source=pr1, tag=int(base_tag + "123")))
        q_dict_waits[col][k].append(comm.irecv(source=pr1, tag=int(base_tag + "1234")))
        q_dict_waits[col][k].append(key[0])


def __split0_q_loop(
    col: int,
    r_tiles: SquareDiagTiles,
    proc_tile_start: torch.Tensor,
    active_procs: torch.Tensor,
    q0_tiles: SquareDiagTiles,
    q_dict: Dict,
    q_dict_waits: Dict,
) -> None:
    """
    Function for Calculating Q for ``split=0`` for QR. ``col`` is the index of the tile column.
    The assumption here is that the diagonal tile is ``(col, col)``.

    Parameters
    ----------
    col : int
        Current column for which to calculate Q
    r_tiles : SquareDiagTiles
        R tiles
    proc_tile_start : torch.Tensor
        Tensor containing the row tile start indices for each process
    active_procs : torch.Tensor
        Tensor containing the ranks of processes with have data
    q0_tiles : SquareDiagTiles
        Q0 tiles
    q_dict : Dict
        Dictionary created in the ``split=0`` R calculation containing all of the Q matrices found
        transforming the matrix to upper triangular for each column. The keys of this dictionary are
        the column indices
    q_dict_waits : Dict
        Dictionary created while sending the Q matrices to the diagonal process
    """
    tile_columns = r_tiles.tile_columns
    diag_process = (
        torch.nonzero(input=proc_tile_start > col, as_tuple=False)[0]
        if col != tile_columns
        else proc_tile_start[-1]
    )
    diag_process = diag_process.item()
    rank = r_tiles.arr.comm.rank
    q0_dtype = q0_tiles.arr.dtype
    q0_torch_type = q0_dtype.torch_type()
    q0_torch_device = q0_tiles.arr.device.torch_device
    # wait for Q tensors sent during the R calculation -----------------------------------------
    if col in q_dict_waits.keys():
        for key in q_dict_waits[col].keys():
            new_key = q_dict_waits[col][key][3] + key + "e"
            q_dict_waits[col][key][0][1].Wait()
            q_dict[col][new_key] = [
                q_dict_waits[col][key][0][0],
                q_dict_waits[col][key][1].wait(),
                q_dict_waits[col][key][2].wait(),
            ]
        del q_dict_waits[col]
    # local Q calculation =====================================================================
    if col in q_dict.keys():
        lcl_col_shape = r_tiles.local_get(key=(slice(None), col)).shape
        # get the start and stop of all local tiles
        #   -> get the rows_per_process[rank] and the row_indices
        row_ind = r_tiles.row_indices
        prev_rows_per_pr = sum(r_tiles.tile_rows_per_process[:rank])
        rows_per_pr = r_tiles.tile_rows_per_process[rank]
        if rows_per_pr == 1:
            # if there is only one tile on the process: return q_dict[col]['0']
            base_q = q_dict[col]["l0"][0].clone()
            del q_dict[col]["l0"]
        else:
            # 0. get the offset of the column start
            offset = (
                torch.tensor(
                    row_ind[col].item() - row_ind[prev_rows_per_pr].item(), device=q0_torch_device
                )
                if row_ind[col].item() > row_ind[prev_rows_per_pr].item()
                else torch.tensor(0, device=q0_torch_device)
            )
            # 1: create an eye matrix of the row's zero'th dim^2
            q_lcl = q_dict[col]["l0"]  # [0] -> q, [1] -> shape of a use in q calc (q is square)
            del q_dict[col]["l0"]
            base_q = torch.eye(
                lcl_col_shape[r_tiles.arr.split], dtype=q_lcl[0].dtype, device=q0_torch_device
            )
            # 2: set the area of the eye as Q
            base_q[offset : offset + q_lcl[1][0], offset : offset + q_lcl[1][0]] = q_lcl[0]

        local_merge_q = {rank: [base_q, None]}
    else:
        local_merge_q = {}
    # -------------- send local Q to all -------------------------------------------------------
    for pr in range(diag_process, active_procs[-1] + 1):
        if pr != rank:
            hld = torch.zeros(
                [q0_tiles.lshape_map[pr][q0_tiles.arr.split]] * 2,
                dtype=q0_torch_type,
                device=q0_torch_device,
            )
        else:
            hld = local_merge_q[pr][0].clone()
        wait = q0_tiles.arr.comm.Ibcast(hld, root=pr)
        local_merge_q[pr] = [hld, wait]

    # recv local Q + apply local Q to Q0
    for pr in range(diag_process, active_procs[-1] + 1):
        if local_merge_q[pr][1] is not None:
            # receive q from the other processes
            local_merge_q[pr][1].Wait()
        if rank in active_procs:
            sum_row = sum(q0_tiles.tile_rows_per_process[:pr])
            end_row = q0_tiles.tile_rows_per_process[pr] + sum_row
            # slice of q_tiles -> [0: -> end local, 1: start -> stop]
            q_rest_loc = q0_tiles.local_get(key=(slice(None), slice(sum_row, end_row)))
            # apply the local merge to q0 then update q0`
            q_rest_loc = q_rest_loc @ local_merge_q[pr][0]
            q0_tiles.local_set(key=(slice(None), slice(sum_row, end_row)), value=q_rest_loc)
            del local_merge_q[pr]

    # global Q calculation =====================================================================
    # split up the Q's from the global QR calculation and set them in a dict w/ proper keys
    global_merge_dict = (
        __split0_global_q_dict_set(
            q_dict_col=q_dict[col], col=col, r_tiles=r_tiles, q_tiles=q0_tiles
        )
        if rank == diag_process
        else {}
    )

    if rank == diag_process:
        merge_dict_keys = set(global_merge_dict.keys())
    else:
        merge_dict_keys = None
    merge_dict_keys = r_tiles.arr.comm.bcast(merge_dict_keys, root=diag_process)

    # send the global merge dictionary to all processes
    for k in merge_dict_keys:
        if rank == diag_process:
            snd = global_merge_dict[k].clone()
            snd_shape = snd.shape
            r_tiles.arr.comm.bcast(snd_shape, root=diag_process)
        else:
            snd_shape = None
            snd_shape = r_tiles.arr.comm.bcast(snd_shape, root=diag_process)
            snd = torch.empty(snd_shape, dtype=q0_dtype.torch_type(), device=q0_torch_device)

        wait = r_tiles.arr.comm.Ibcast(snd, root=diag_process)
        global_merge_dict[k] = [snd, wait]
    if rank in active_procs:
        # create a dictionary which says what tiles are in each column of the global merge Q
        qi_mult = {}
        for c in range(q0_tiles.tile_columns):
            # this loop is to slice the merge_dict keys along each column + create the
            qi_mult_set = set([(i, c) for i in range(col, q0_tiles.tile_columns)])
            if len(qi_mult_set & merge_dict_keys) != 0:
                qi_mult[c] = list(qi_mult_set & merge_dict_keys)

        # have all the q_merge in one place, now just do the mm with q0
        # get all the keys which are in a column (qi_mult[column])
        row_inds = q0_tiles.row_indices + [q0_tiles.arr.gshape[0]]
        q_copy = q0_tiles.arr.larray.clone()
        for qi_col in qi_mult.keys():
            # multiply q0 rows with qi cols
            # the result of this will take the place of the row height and the column width
            out_sz = q0_tiles.local_get(key=(slice(None), qi_col)).shape
            mult_qi_col = torch.zeros(
                (q_copy.shape[1], out_sz[1]), dtype=q0_dtype.torch_type(), device=q0_torch_device
            )
            for ind in qi_mult[qi_col]:
                if global_merge_dict[ind][1] is not None:
                    global_merge_dict[ind][1].Wait()
                lp_q = global_merge_dict[ind][0]
                if mult_qi_col.shape[1] < lp_q.shape[1]:
                    new_mult = torch.zeros(
                        (mult_qi_col.shape[0], lp_q.shape[1]),
                        dtype=mult_qi_col.dtype,
                        device=q0_torch_device,
                    )
                    new_mult[:, : mult_qi_col.shape[1]] += mult_qi_col.clone()
                    mult_qi_col = new_mult

                mult_qi_col[
                    row_inds[ind[0]] : row_inds[ind[0]] + lp_q.shape[0], : lp_q.shape[1]
                ] = lp_q
            hold = torch.matmul(q_copy, mult_qi_col)

            write_inds = q0_tiles.get_start_stop(key=(0, qi_col))
            q0_tiles.arr.lloc[:, write_inds[2] : write_inds[2] + hold.shape[1]] = hold
    else:
        for ind in merge_dict_keys:
            global_merge_dict[ind][1].Wait()
    if col in q_dict.keys():
        del q_dict[col]


def __split1_qr_loop(
    dcol: int, r_tiles: SquareDiagTiles, q0_tiles: SquareDiagTiles, calc_q: bool
) -> None:
    """
    Helper function to do the QR factorization of the column ``dcol``.
    This function assumes that the target tile is at ``(dcol, dcol)``. This is the standard case at it assumes that
    the diagonal tile holds the diagonal entries of the matrix.

    Parameters
    ----------
    dcol : int
        Column of the diagonal process
    r_tiles : SquareDiagTiles
        Input matrix tiles to QR,
        if copy is ``True`` in QR then it is a copy of the data, else it is the same as the input
    q0_tiles : SquareDiagTiles
        The Q matrix tiles as created in the QR function.
    calc_q : bool
        Flag for weather to calculate Q or not, if ``False``, then ``Q=None``

    """
    r_torch_device = r_tiles.arr.larray.device
    q0_torch_device = q0_tiles.arr.larray.device if calc_q else None
    # ==================================== R Calculation - single tile =========================
    # loop over each column, need to do the QR for each tile in the column(should be rows)
    # need to get the diagonal process
    rank = r_tiles.arr.comm.rank
    cols_on_proc = torch.cumsum(
        torch.tensor(r_tiles.tile_columns_per_process, device=r_torch_device), dim=0
    )
    not_completed_processes = torch.nonzero(input=dcol < cols_on_proc, as_tuple=False).flatten()
    diag_process = not_completed_processes[0].item()
    tile_rows = r_tiles.tile_rows
    # get the diagonal tile and do qr on it
    # send q to the other processes
    # 1st qr: only on diagonal tile + apply to the row
    if rank == diag_process:
        # do qr on diagonal process
        try:
            q1, r1 = torch.linalg.qr(r_tiles[dcol, dcol], mode="complete")
        except AttributeError:
            q1, r1 = r_tiles[dcol, dcol].qr(some=False)

        r_tiles.arr.comm.Bcast(q1.clone(), root=diag_process)
        r_tiles[dcol, dcol] = r1
        # apply q1 to the trailing matrix (other processes)

        # need to convert dcol to a local index
        loc_col = dcol - sum(r_tiles.tile_columns_per_process[:rank])
        hold = r_tiles.local_get(key=(dcol, slice(loc_col + 1, None)))
        if hold is not None:  # if there is more data on that row after the diagonal tile
            r_tiles.local_set(key=(dcol, slice(loc_col + 1, None)), value=torch.matmul(q1.T, hold))
    elif rank > diag_process:
        # recv the Q from the diagonal process, and apply it to the trailing matrix
        st_sp = r_tiles.get_start_stop(key=(dcol, dcol))
        sz = st_sp[1] - st_sp[0], st_sp[3] - st_sp[2]

        q1 = torch.zeros(
            (sz[0], sz[0]), dtype=r_tiles.arr.dtype.torch_type(), device=r_torch_device
        )
        loc_col = 0
        r_tiles.arr.comm.Bcast(q1, root=diag_process)
        hold = r_tiles.local_get(key=(dcol, slice(0, None)))
        r_tiles.local_set(key=(dcol, slice(0, None)), value=torch.matmul(q1.T, hold))
    else:
        # these processes are already done calculating R, only need to calc Q, need to recv q1
        st_sp = r_tiles.get_start_stop(key=(dcol, dcol))
        sz = st_sp[1] - st_sp[0], st_sp[3] - st_sp[2]
        q1 = torch.zeros(
            (sz[0], sz[0]), dtype=r_tiles.arr.dtype.torch_type(), device=r_torch_device
        )
        r_tiles.arr.comm.Bcast(q1, root=diag_process)

    # ================================ Q Calculation - single tile =============================
    if calc_q:
        for row in range(q0_tiles.tile_rows_per_process[rank]):
            # q1 is applied to each tile of the column dcol of q0 then written there
            q0_tiles.local_set(
                key=(row, dcol), value=torch.matmul(q0_tiles.local_get(key=(row, dcol)), q1)
            )
    del q1
    # loop over the rest of the rows, combine the tiles, then apply the result to the rest
    # 2nd step: merged QR on the rows
    # ================================ R Calculation - merged tiles ============================
    diag_tile = r_tiles[dcol, dcol]
    # st_sp = r_tiles.get_start_stop(key=(dcol, dcol))
    diag_st_sp = r_tiles.get_start_stop(key=(dcol, dcol))
    diag_sz = diag_st_sp[1] - diag_st_sp[0], diag_st_sp[3] - diag_st_sp[2]
    # (Q) need to get the start stop of diag tial
    for row in range(dcol + 1, tile_rows):
        lp_st_sp = r_tiles.get_start_stop(key=(row, dcol))
        lp_sz = lp_st_sp[1] - lp_st_sp[0], lp_st_sp[3] - lp_st_sp[2]
        if rank == diag_process:
            # cat diag tile and loop tile
            loop_tile = r_tiles[row, dcol]
            loop_cat = torch.cat((diag_tile, loop_tile), dim=0)
            # qr
            try:
                ql, rl = torch.linalg.qr(loop_cat, mode="complete")
            except AttributeError:
                ql, rl = loop_cat.qr(some=False)
            # send ql to all
            r_tiles.arr.comm.Bcast(ql.clone().contiguous(), root=diag_process)
            # set rs
            r_tiles[dcol, dcol] = rl[: diag_sz[0]]
            r_tiles[row, dcol] = rl[diag_sz[0] :]
            # apply q to rest
            if loc_col + 1 < r_tiles.tile_columns_per_process[rank]:
                upp = r_tiles.local_get(key=(dcol, slice(loc_col + 1, None)))
                low = r_tiles.local_get(key=(row, slice(loc_col + 1, None)))
                hold = torch.matmul(ql.T, torch.cat((upp, low), dim=0))
                # set upper
                r_tiles.local_set(key=(dcol, slice(loc_col + 1, None)), value=hold[: diag_sz[0]])
                # set lower
                r_tiles.local_set(key=(row, slice(loc_col + 1, None)), value=hold[diag_sz[0] :])
        elif rank > diag_process:
            ql = torch.zeros(
                [lp_sz[0] + diag_sz[0]] * 2,
                dtype=r_tiles.arr.dtype.torch_type(),
                device=r_torch_device,
            )
            r_tiles.arr.comm.Bcast(ql, root=diag_process)
            upp = r_tiles.local_get(key=(dcol, slice(0, None)))
            low = r_tiles.local_get(key=(row, slice(0, None)))
            hold = torch.matmul(ql.T, torch.cat((upp, low), dim=0))
            # set upper
            r_tiles.local_set(key=(dcol, slice(0, None)), value=hold[: diag_sz[0]])
            # set lower
            r_tiles.local_set(key=(row, slice(0, None)), value=hold[diag_sz[0] :])
        else:
            ql = torch.zeros(
                [lp_sz[0] + diag_sz[0]] * 2,
                dtype=r_tiles.arr.dtype.torch_type(),
                device=r_torch_device,
            )
            r_tiles.arr.comm.Bcast(ql, root=diag_process)
        # ================================ Q Calculation - merged tiles ========================
        if calc_q:
            top_left = ql[: diag_sz[0], : diag_sz[0]]
            top_right = ql[: diag_sz[0], diag_sz[0] :]
            bottom_left = ql[diag_sz[0] :, : diag_sz[0]]
            bottom_right = ql[diag_sz[0] :, diag_sz[0] :]
            # two multiplications: one for the left tiles and one for the right
            # left tiles --------------------------------------------------------------------
            # create r column of the same size as the tile row of q0
            st_sp = r_tiles.get_start_stop(key=(slice(dcol, None), dcol))
            qloop_col_left_sz = st_sp[1] - st_sp[0], st_sp[3] - st_sp[2]
            qloop_col_left = torch.zeros(
                qloop_col_left_sz, dtype=q0_tiles.arr.dtype.torch_type(), device=q0_torch_device
            )
            # top left starts at 0 and goes until diag_sz[1]
            qloop_col_left[: diag_sz[0]] = top_left
            # bottom left starts at ? and goes until ? (only care about 0th dim)
            st, sp, _, _ = r_tiles.get_start_stop(key=(row, 0))
            st -= diag_st_sp[0]  # adjust these by subtracting the start index of the diag tile
            sp -= diag_st_sp[0]
            qloop_col_left[st:sp] = bottom_left
            # right tiles --------------------------------------------------------------------
            # create r columns tensor of the size of the tile column of index 'row'
            st_sp = q0_tiles.get_start_stop(key=(row, slice(dcol, None)))
            sz = st_sp[1] - st_sp[0], st_sp[3] - st_sp[2]
            qloop_col_right = torch.zeros(
                sz[1], sz[0], dtype=q0_tiles.arr.dtype.torch_type(), device=q0_torch_device
            )
            # top left starts at 0 and goes until diag_sz[1]
            qloop_col_right[: diag_sz[0]] = top_right
            # bottom left starts at ? and goes until ? (only care about 0th dim)
            st, sp, _, _ = r_tiles.get_start_stop(key=(row, 0))
            st -= diag_st_sp[0]  # adjust these by subtracting the start index of the diag tile
            sp -= diag_st_sp[0]
            qloop_col_right[st:sp] = bottom_right
            for qrow in range(q0_tiles.tile_rows_per_process[rank]):
                # q1 is applied to each tile of the column dcol of q0 then written there
                q0_row = q0_tiles.local_get(key=(qrow, slice(dcol, None))).clone()
                q0_tiles.local_set(key=(qrow, dcol), value=torch.matmul(q0_row, qloop_col_left))
                q0_tiles.local_set(key=(qrow, row), value=torch.matmul(q0_row, qloop_col_right))
        del ql



"""
Module for Spectral Clustering, a graph-based machine learning algorithm
"""

import heat as ht
import math
import torch
from typing import Tuple
from heat.core.dndarray import DNDarray


class Spectral(ht.ClusteringMixin, ht.BaseEstimator):
    """
    Spectral clustering

    Attributes
    ----------
    n_clusters : int
        Number of clusters to fit
    gamma : float
        Kernel coefficient sigma for 'rbf', ignored for metric='euclidean'
    metric : string
        How to construct the similarity matrix.

            - 'rbf' : construct the similarity matrix using a radial basis function (RBF) kernel.
            - 'euclidean' : construct the similarity matrix as only euclidean distance.
    laplacian : str
        How to calculate the graph laplacian (affinity)
        Currently supported : 'fully_connected', 'eNeighbour'
    threshold : float
        Threshold for affinity matrix if laplacian='eNeighbour'
        Ignorded for laplacian='fully_connected'
    boundary : str
        How to interpret threshold: 'upper', 'lower'
        Ignorded for laplacian='fully_connected'
    n_lanczos : int
        number of Lanczos iterations for Eigenvalue decomposition
    assign_labels: str
         The strategy to use to assign labels in the embedding space.
    **params: dict
          Parameter dictionary for the assign_labels estimator
    """

    def __init__(
        self,
        n_clusters: int = None,
        gamma: float = 1.0,
        metric: str = "rbf",
        laplacian: str = "fully_connected",
        threshold: float = 1.0,
        boundary: str = "upper",
        n_lanczos: int = 300,
        assign_labels: str = "kmeans",
        **params,
    ):
        self.n_clusters = n_clusters
        self.gamma = gamma
        self.metric = metric
        self.laplacian = laplacian
        self.threshold = threshold
        self.boundary = boundary
        self.n_lanczos = n_lanczos
        self.assign_labels = assign_labels

        if metric == "rbf":
            sig = math.sqrt(1 / (2 * gamma))
            self._laplacian = ht.graph.Laplacian(
                lambda x: ht.spatial.rbf(x, sigma=sig, quadratic_expansion=True),
                definition="norm_sym",
                mode=laplacian,
                threshold_key=boundary,
                threshold_value=threshold,
            )

        elif metric == "euclidean":
            self._laplacian = ht.graph.Laplacian(
                lambda x: ht.spatial.cdist(x, quadratic_expansion=True),
                definition="norm_sym",
                mode=laplacian,
                threshold_key=boundary,
                threshold_value=threshold,
            )
        else:
            raise NotImplementedError("Other kernels currently not supported")

        if assign_labels == "kmeans":
            self._cluster = ht.cluster.KMeans(params)
        else:
            raise NotImplementedError(
                "Other Label Assignment Algorithms are currently not available"
            )

        # in-place properties
        self._labels = None

    @property
    def labels_(self) -> DNDarray:
        """
        Returns labels of each point.
        """
        return self._labels

    def _spectral_embedding(self, x: DNDarray) -> Tuple[DNDarray, DNDarray]:
        """
        Helper function for dataset x embedding.
        Returns Tupel(Eigenvalues, Eigenvectors) of the graph's Laplacian matrix.

        Parameters
        ----------
        x : DNDarray
            Sample Matrix for which the embedding should be calculated

        Notes
        -----
        This will throw out the complex side of the eigenvalues found during this.

        """
        L = self._laplacian.construct(x)
        # 3. Eigenvalue and -vector calculation via Lanczos Algorithm
        v0 = ht.full(
            (L.shape[0],),
            fill_value=1.0 / math.sqrt(L.shape[0]),
            dtype=L.dtype,
            split=0,
            device=L.device,
        )
        V, T = ht.lanczos(L, self.n_lanczos, v0)

        # if int(torch.__version__.split(".")[1]) >= 9:
        try:
            # 4. Calculate and Sort Eigenvalues and Eigenvectors of tridiagonal matrix T
            eval, evec = torch.linalg.eig(T.larray)

            # If x is an Eigenvector of T, then y = V@x is the corresponding Eigenvector of L
            eval, idx = torch.sort(eval.real, dim=0)
            eigenvalues = ht.array(eval)
            eigenvectors = ht.matmul(V, ht.array(evec))[:, idx]

            return eigenvalues.real, eigenvectors.real
        except AttributeError:  # torch version is less than 1.9.0
            # 4. Calculate and Sort Eigenvalues and Eigenvectors of tridiagonal matrix T
            eval, evec = torch.eig(T.larray, eigenvectors=True)
            # If x is an Eigenvector of T, then y = V@x is the corresponding Eigenvector of L
            eval, idx = torch.sort(eval[:, 0], dim=0)
            eigenvalues = ht.array(eval)
            eigenvectors = ht.matmul(V, ht.array(evec))[:, idx]

            return eigenvalues, eigenvectors

    def fit(self, x: DNDarray):
        """
        Clusters dataset X via spectral embedding.
        Computes the low-dim representation by calculation of eigenspectrum (eigenvalues and eigenvectors) of the graph
        laplacian from the similarity matrix and fits the eigenvectors that correspond to the k lowest eigenvalues with
        a seperate clustering algorithm (currently only kmeans is supported). Similarity metrics for adjacency
        calculations are supported via spatial.distance. The eigenvalues and eigenvectors are computed by reducing the
        Laplacian via lanczos iterations and using the torch eigenvalue solver on this smaller matrix. If other
        eigenvalue decompostion methods are supported, this will be expanded.

        Parameters
        ----------
        x : DNDarray
            Training instances to cluster. Shape = (n_samples, n_features)
        """
        # 1. input sanitation
        if not isinstance(x, DNDarray):
            raise ValueError(f"input needs to be a ht.DNDarray, but was {type(x)}")
        if x.split is not None and x.split != 0:
            raise NotImplementedError("Not implemented for other splitting-axes")
        # 2. Embed Dataset into lower-dimensional Eigenvector space
        eigenvalues, eigenvectors = self._spectral_embedding(x)

        # 3. Find the spectral gap, if number of clusters is not defined from the outside
        if self.n_clusters is None:
            diff = eigenvalues[1:] - eigenvalues[:-1]
            tmp = ht.argmax(diff).item()
            self.n_clusters = tmp + 1

        components = eigenvectors[:, : self.n_clusters].copy()

        params = self._cluster.get_params()
        params["n_clusters"] = self.n_clusters
        self._cluster.set_params(**params)
        self._cluster.fit(components)
        self._labels = self._cluster.labels_
        self._cluster_centers = self._cluster.cluster_centers_

        return self

    def predict(self, x: DNDarray) -> DNDarray:
        """
        Return the label each sample in X belongs to.
        X is transformed to the low-dim representation by calculation of eigenspectrum (eigenvalues and eigenvectors) of
        the graph laplacian from the similarity matrix. Inference of lables is done by extraction of the closest
        centroid of the n_clusters eigenvectors from the previously fitted clustering algorithm (kmeans).

        Parameters
        ----------
        x : DNDarray
            New data to predict. Shape = (n_samples, n_features)

        Warning
        -------
        Caution: Calculation of the low-dim representation requires some time!

        """
        # input sanitation
        if not isinstance(x, DNDarray):
            raise ValueError(f"input needs to be a ht.DNDarray, but was {type(x)}")
        if x.split is not None and x.split != 0:
            raise NotImplementedError("Not implemented for other splitting-axes")

        _, eigenvectors = self._spectral_embedding(x)

        components = eigenvectors[:, : self.n_clusters].copy()

        return self._cluster.predict(components)
